import datetime
import inspect
import json
import os
import shutil
import sys
import time
import bragg_objects

from scipy.interpolate import interp1d as interp
from tmmpy import *

# A little Python magic to resolve current location (used for relative-path resource loading)
core_dir = os.path.dirname(os.path.abspath(inspect.stack()[0][1]))


# region Progress printing and control flow

def dispatch(name, calc_type, funcs, wd=None, cores=None, clean=True):
    """
    To avoid writing the calculation dispatch logic a million times (or, well three times so far) all dispatching logic
    is implemented here. It includes information/progress printing, multi-core dispatching logic and so on.
    :param clean: If true, the working directory is cleared before calculation launch
    :param name: calculation name, a folder with the name is created (and cleaned!) to store all calculation files
    :param calc_type: calculation type, used e.g. for info printing
    :param funcs: functional injections, must contain "build_info" (default is no additional info written),
    "build_slices", "build_result" (default is no result building) and "calc"
    :param wd: working direction, the calculation folder is created in this directory
    :param cores: number of cores to utilize, default is to use all cores available to the system
    :return: list of calculation results
    """
    # Log start.
    tick1 = time.time()
    tick_str = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    print("")
    print("***** {} calculation: {} STARTED at {} *****".format(calc_type.upper(), name, tick_str))
    # Setup working directory.
    cwd = name if wd is None else os.path.join(wd, name)
    if os.path.exists(cwd) and clean:
        shutil.rmtree(cwd, True)
    if not os.path.exists(cwd):
        os.makedirs(cwd)
    # Write info to file.
    info = {} if 'build_info' not in funcs else funcs['build_info']()
    info["calc"] = funcs["calc"].__name__
    info["time"] = tick_str
    info["cores"] = cores
    info["data_version"] = bragg_objects.__data_version__
    with open(os.path.join(cwd, "{}.info".format(calc_type.lower())), "w") as f:
        json.dump(info, f)
    # Generate slices.
    slice_map = funcs['build_slices'](cwd)
    slices = slice_map["slices"]
    cores = min(cores, len(slices))
    # Log calculation launch.
    dt = time.time() - tick1
    print("Pre processing took {:.2f}s. ".format(dt), end='')
    print("Identified {} slices, spawning {} process(es). \n".format(len(slices), cores))
    in_ide = any('PYCHARM' in name for name in os.environ)
    prg_bar = MyProgressBar(len(slices), delay=5 if not in_ide else 1 - 3)
    # Currently only one core is supported for Windows, i.e. freeze it.
    cores = 1 if "Windows" in platform.system() else cores
    # Set process affinity (fixed problem with numpy processes only assigned to one core).
    if cores is not 1:
        os.system('taskset -cp 0-%d %s' % (cores, os.getpid()))
    # Launch calculations.
    entries = []
    if cores is 1:
        for i in range(len(slices)):
            entries.append(funcs["calc"](slices[i]))
            prg_bar.update(i + 1)
    else:
        pool = multiprocessing.Pool(cores)
        for i, result in enumerate(pool.imap(funcs["calc"], slices), 1):
            entries.append(result)
            prg_bar.update(i + 1)
        pool.close()
    # Collect results into result object.
    print("")
    result = None if 'build_result' not in funcs else funcs["build_result"](cwd, slice_map, entries)
    # Log calculation ended.
    toc11 = time.time()
    toc_str = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    print("Calculation took {:.2f}s.".format((toc11 - tick1)))
    print("***** {} calculation: {} ENDED at {} *****".format(calc_type.upper(), name, toc_str))
    print("")

    return result


class MyProgressBar:
    def __init__(self, total, steps=None, delay=None):
        """
        Convenient wrapper of the print_progress method which holds state information
        @params:
            total       - Required  : total iterations (Int)
            steps       - Optional  : steps at which progress is printed, not None => update on new lines (list)
            delay       - Optional  : not None => update only printed every delay-th second
        """
        self.total = total
        self.steps = steps
        self.delay = 0 if delay is None else delay
        self.step = 0
        self.tick = time.time()
        self.last = time.time()
        self.done = False

    def update(self, iteration):
        """
        Update progress bar.
        :param iteration: current iteration
        :return: None
        """
        prev = self.last
        now = time.time()
        # If no steps are specified, print every delay-th time.
        if (now - prev) < self.delay and self.steps is None:
            return
        self.last = now
        # Never print 100% twice.
        if self.done:
            return
        # Print a each step or every time if no steps are specified.
        pct = 100 * (iteration / float(self.total))
        if self.steps is None or pct > self.steps[self.step]:
            self.step += 1
            etc = self.tick + (self.last - self.tick) / iteration * self.total
            etc_str = datetime.datetime.fromtimestamp(etc).strftime('%Y-%m-%d %H:%M:%S')
            # When steps are specified, each update is written to a new line.
            prefix = '\r' if self.steps is None else ""
            suffix = "(ETC {})".format(etc_str)
            itr = min(iteration, self.total)
            print_progress(itr, self.total, prefix=prefix, suffix=suffix)
            self.done = itr == self.total


# From https://gist.github.com/aubricus/f91fb55dc6ba5557fbab06119420dd6a
def print_progress(iteration, total, prefix='', suffix='', decimals=1, bar_length=30):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        bar_length  - Optional  : character length of bar (Int)
    """
    str_format = "{0:5." + str(decimals) + "f}"
    percents = str_format.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    bar = '=' * filled_length + '-' * (bar_length - filled_length)

    sys.stdout.write('%s |%s| %s%s %s' % (prefix, bar, percents, '%', suffix)),

    if iteration == total:
        sys.stdout.write('\n')
    sys.stdout.flush()


# endregion

# region Export logic

def origin_export1d(x, vector, name, wd=None):
    """
    Export data in origin-friendly format.
    :param wd:
    :param x: x header values
    :param y: y header values
    :param vector: matrix data
    :param name: file name (without extension) to which the data should be saved
    :return: None
    """
    if len(vector.shape) is not 1:
        raise ValueError("Data must be one-dimensional for origin_export1d.")
    # Create Origin data.
    result = np.zeros(vector.shape[0], 2)
    result[:, 0] = x
    result[:, 1] = vector
    # Save the file.
    file_name = "{}.txt".format(name)
    file_path = os.path.join(wd, file_name) if wd is not None else file_name
    np.savetxt(file_path, result)


def origin_export2d(x, y, matrix, name, wd=None):
    """
    Export data in origin-friendly format.
    :param wd:
    :param x: x header values
    :param y: y header values
    :param matrix: matrix data
    :param name: file name (without extension) to which the data should be saved
    :return: None
    """
    if len(matrix.shape) is not 2:
        raise ValueError("Data must be two-dimensional for origin_export2d.")
    # Create Origin data.
    result = np.zeros((matrix.shape[0] + 1, matrix.shape[1] + 1))
    result[1:, 0] = x
    result[0, 1:] = y
    result[1:, 1:] = matrix
    # Save the file.
    file_name = "{}.txt".format(name)
    file_path = os.path.join(wd, file_name) if wd is not None else file_name
    np.savetxt(file_path, result)


# endregion

# region Resource logic

def get_resource_path(res):
    """
    Get resource path.
    :param res: resource name (must correspond to a file name in the resource folder)
    :return: absolute resource path
    """
    return os.path.join(core_dir, "res/{}".format(res))


def read_material(mat_key, zero_abs=True):
    """
    Read material from (internal) resources.
    :param mat_key: material spectrum identifier
    :param zero_abs: set absorption to zero (default behaviour)
    :return: material object
    """
    mat_map = {"pmma_np": "generated_nk_data_PMMA_NP_CANdots.dat", "pmma": "generated_nk_data_PMMA_withoutNP.dat",
               "tio2": "nk_TiO2_tempmean.dat", "tio2_ald": "nk_TiO2_ALD_k0_350_2000.dat",
               "17_tio2_ald": "1708_si_sio2_tio2_codyLorentz_roughness.dat",
               "1711_tio2": "1711_si_sio2_tio2_cody_lorentz_FINAL_nk_250_2000.dat",
               "npsf": "nk_PMMA_NPSF_cauchy_250_2000.dat",
               "17_npsf": "1708_si_sio2_pmma_np_cauchy_abs_270.dat",
               "1711_pmmanp": "1711_si_sio2_pmmanp_cauchy_FINAL_nk_250_2000.dat"}
    data = np.loadtxt(get_resource_path(mat_map[mat_key]))
    ri = TabulatedIndex(data[:, 0], data[:, 1])
    ec = 0 if zero_abs else TabulatedIndex(data[:, 0], data[:, 2])
    return Material(ri, ec, name=mat_key)


def read_absorption_spectrum(abs_key):
    """
    Read absorption spectrum from (internal) resources.
    :param abs_key: absorption spectrum identifier
    :return: absorption spectrum interpolation object
    """
    abs_map = {"abs_er": "Norm_Absorption_NaYF4_25Er_170-230mesh_1375_1650.dat"}
    data = np.loadtxt(get_resource_path(abs_map[abs_key]), skiprows=1)
    return interp(data[:, 0], data[:, 1])

# endregion
