import numpy as np
from numpy import cos, sin, arcsin

def dos3d_analytic(n_int, dk, w_bins):
    """
    Calculate the binned 3D DOS analytically for a homogeneous medium.
    :param n_int: refractive index of internal medium (reference, typically pmma)
    :param dk: bin spacing in k-space
    :param w_bins: vector of omega bins
    :return: vector of dos values, one per frequency bin
    """
    dw = w_bins[1] - w_bins[0]
    dos3d = np.zeros_like(w_bins)
    for i, w in enumerate(w_bins):
        # Formula based on (3.13) + technique like (3.14) in Gutmann thesis. Note than here we count from w to w + dw.
        dos3d[i] = 4 * np.pi * n_int ** 3 * dw / dk ** 2 * (w ** 2 + w * dw + dw ** 2 / 3)
    return dos3d

def fdos3d_phi_d_analytic(phi_ds, theta_d, n_int, n_out):
    """
    Calculate angular distribution of fractional 3D DOS analytically for a homogeneous medium. In this special case
    the angular distribution is independent of frequency.
    :param phi_ds: vector of detection angles in degrees
    :param theta_d: detection cone half-angle in degrees
    :param n_int: refractive index of internal medium (reference, typically pmma)
    :param n_out: refractive index of external medium (typically air)
    :return: vector of dos3d weighting factors, one per collection angle
    """
    fdos3d_ratios = np.zeros_like(phi_ds, dtype=float)
    # Setup uniform theta values on xy-plane.
    emission_ang = np.linspace(0, np.pi / 2, 100000)
    ky = np.array(sin(emission_ang) * n_int)
    norm = 2 * np.pi * np.sum(ky)
    for i, phi_d in enumerate(phi_ds):
        # Calculate which modes escape (Gutmann, page 33).
        escaped = ky <= n_out
        # Calculate which modes that are collected at the angle phi_ḱ (Gutmann, 3.22)
        phi_k = np.empty_like(ky)
        phi_k.fill(180)  # dummy value, indicates that the mode is NOT collected
        phi_k[escaped] = np.rad2deg(np.arcsin(ky[escaped] / n_out))
        collected = np.logical_and(phi_d - theta_d <= phi_k, phi_k <= phi_d + theta_d)
        # Calculate the detection fraction.
        fd = np.zeros_like(ky)
        fd[collected] = frac_detectable(theta_d, phi_d, 1, ky[collected])
        fdos3d_ratios[i] = np.nansum(2 * np.pi * ky * fd) / norm
    return fdos3d_ratios

def fdos3d_omega_phi_d_analytic(n_int, n_out, dk, w_bins, phi_ds, theta_d):
    """
    Calculate the fractional 3D DOS on a omega bins/phi_d angle grid analytically for a homogeneous medium.
    :param n_int: refractive index of internal medium (typically pmma)
    :param n_out: refractive index of external medium (typically air)
    :param dk: bin spacing in k-space
    :param w_bins: vector of omega bins
    :param phi_ds: vector of detection angles in degrees
    :param theta_d: detection cone half-angle in degrees
    :return: matrices of bins/angles/FDOS3D
    """
    detection_fracs = fdos3d_phi_d_analytic(phi_ds, theta_d, n_int, n_out)
    dos3d = dos3d_analytic(n_int, dk, w_bins)
    X, Y = np.meshgrid(w_bins, phi_ds)
    Xv, Yv = np.meshgrid(dos3d, detection_fracs)
    Z = Xv * Yv
    return X, Y, Z

def frac_detectable(theta_d, phi_d, kn, ky):
    """
    Calculate (analytically) the detectable fraction.
    :param theta_d: detection cone half-angle in degrees
    :param phi_d: collection angle in degrees
    :param kn: length of k-vector (same a frequency in dimensionless form)
    :param ky: y-component of k-vector
    :return:
    """
    # Angles are assumed in degrees, convert them.
    phi_d = np.deg2rad(phi_d)
    theta_d = np.deg2rad(theta_d)
    # At normal incidence, the collection angle is "all".
    if phi_d == 0:
        return 1
    # Analytic solution for the detection fraction, see examples for illustration.
    phi_k = arcsin(ky / kn)
    frac = (cos(theta_d) - cos(phi_d) * cos(phi_k)) / (sin(phi_d) * sin(phi_k))
    return np.arccos(frac).real / np.pi