import subprocess
import h5py

from bragg_objects import *
from bragg_utils import *

core_dir = os.path.dirname(os.path.abspath(inspect.stack()[0][1]))


# region Interface to mpb (external calls, slicing etc.)


def calculate_mpb_result(args):
    """
    Calculation of mpb result from a prepared args set (or slice). Should never be called directly, but only
    through the main entry point 'run_mpb_calculation'. As the calculation is isolated to the assigned wd, many
    calculations can be run at once.
    :param args: arguments should include "name", "wd" and any parameters to be passed to mpb (see "build_ctl_script")
    :return: working directory, mpb result
    """
    # Setup the folder structure.
    wd = args["wd"]
    if os.path.exists(wd):
        shutil.rmtree(wd)
    os.makedirs(wd)
    # Setup the ctl script.
    os.environ['GUILE_WARN_DEPRECATED'] = "no"
    ctl_script = build_ctl_script(args)
    with open(os.path.join(wd, '{}.ctl'.format(args["name"])), 'a') as my_file:
        my_file.write(ctl_script)
    # Launch the calculation.
    shutil.copy2(os.path.join(core_dir, 'run_mpb.sh'), os.path.join(wd, 'run_mpb.sh'))
    subprocess.call(['git-bash.exe' if "Windows" in platform.system() else "bash", 'run_mpb.sh', args["name"]], cwd=wd)
    return wd, build_mpb_result_object(wd, args)


def build_ctl_script(args):
    """
    Build the control script used to run the mpb calculation. A template is used ("template.ctl"), but most parameters
    are injected directly, e.g. written to the ctl script based on the content of args.
    :param args: arguments passed to mpb, should include "n_low", "n_high", "num_bands", "modes", "field" and 
    "resolution"
    :return: ctl script as a string for mpb calculation
    """
    x1, y1, x2, y2 = args['x1'], args['y1'], args['x2'], args['y2']
    interp = "num-ks" in args.keys()
    # Setup grid of k-points.
    script = ['; setup k-points\n']
    if interp:
        # Use the built-in interpolation function from mpb.
        script.append("(set! k-points (list\n")
        script.append("    (vector3 {:.5f} {:.5f} 0)\n".format(x1, y1))
        script.append("    (vector3 {:.5f} {:.5f} 0)))\n".format(x2, y2))
        script.append("(set! k-points (interpolate {} k-points))\n".format(int(args["num-ks"])))
    else:
        # Interpolate k-points directly and write values to ctl file (+1e-15 to force endpoint inclusion).
        kxs = np.arange(x1, x2 + 1e-15, args["dx"]) if x1 != x2 else [x1]
        kys = np.arange(y1, y2 + 1e-15, args["dy"]) if y1 != y2 else [y1]
        script.append("(set! k-points (list\n")
        for i in range(len(kxs)):
            for j in range(len(kys)):
                script.append("    (vector3 {:.5f} {:.5f} 0)\n".format(kxs[i], kys[j]))
        script[-1] = script[-1].replace(")\n", ")))\n")
    # Initialize variables.
    script.append('; injected variables \n')
    to_define = ["n-low", "n-high"]
    already_defined = ["num-bands", "resolution"]
    for key in to_define:
        script.append("(define-param {} {}) \n".format(key, args[key.replace("-", "_")]))
    for key in already_defined:
        if key.replace("-", "_") not in args:
            continue
        script.append("(set-param! {} {}) \n".format(key, args[key.replace("-", "_")]))
    # Load script template.
    with open(os.path.join(core_dir, 'template.ctl'), 'r') as my_file:
        script.append(my_file.read())
    # Add calculations.
    for mode in args["modes"]:
        if "field" in args:
            script.append("(run-{} fix-dfield-phase output-{}) \n".format(mode, args["field"]))
        else:
            script.append("(run-{}) \n".format(mode))
    return ''.join(script)


def build_mpb_result_object(wd, args):
    """
    Build mpb result object from the output of the MPB calculation. Includes cleanup, e.g. the files (many files!)
    output from MPB are deleted immediately.
    :param wd: working directory of mpb calculation
    :param args: arguments for mpb calculation, should contain "sym", "modes", "field" and "resolution"
    :return: mpb result object ("MpbResult")
    """
    sym = "sym" in args and args["sym"]
    # Read epsilon file.
    epsilon_path = os.path.join(wd, "{}-epsilon.h5".format(args["name"]))
    with h5py.File(epsilon_path, 'r') as f:
        epsilon_data = np.array(f["data"])
    # If the field is symmetric, it is enough to save the first half.
    if sym:
        epsilon_data = np.array(epsilon_data[:int(len(epsilon_data) / 2.0) + 1])
    # Read data for each mode.
    ws = {}
    fields = {}
    for mode in args["modes"]:
        # Read band file.
        band_path = os.path.join(wd, "{}freqs.txt".format(mode))
        with open(band_path) as in_file:
            header = in_file.readline().rstrip().split(",")
        data = np.loadtxt(band_path, skiprows=1, delimiter=",", usecols=range(2, len(header)))
        kx, ky, kz, ws[mode] = data[:, 0], data[:, 1], data[:, 2], data[:, 4:]
        # Check if fields are to be recorded.
        if "field" not in args:
            continue
        # Read field files.
        ks = len(kx)
        bs = ws[mode].shape[1]
        fields[mode] = np.empty((ks, bs, len(epsilon_data)))
        for k in range(ks):
            for b in range(bs):
                h5_file = "{}-{}.k{:02d}.b{:02d}.{}.h5".format(args["name"], args["field"], k + 1, b + 1, mode)
                h5_path = os.path.join(wd, h5_file)
                with h5py.File(h5_path, 'r') as f:
                    fields[mode][k, b, :] = np.array(f["data"])[:len(epsilon_data)] / args["resolution"]
    # Clean up the mess.
    shutil.rmtree(wd)
    return MpbResult(kx, ky, kz, epsilon_data, ws, fields, args["sym"])


def dump_mpb_result(args_list):
    """
    Dump mpb result to .json file. Default behaviour of the "run_mpb_calculation". Could just have been a lambda,
    but due to the pickling done by the multiprocessing package, the method MUST be defined on module level.
    :param args_list: list of arguments for mpb calculations (see "calculate_mpb_result").
    :return: None
    """
    for args in args_list:
        wd, mpb_result = calculate_mpb_result(args)
        with open("{}.json".format(wd), 'w+') as out_file:
            out_file.write(jsonpickle.encode(mpb_result))


def agg_mpb_results(wd, results):
    """
    Aggregation of mpb results. As mpb results are large (does not fit in memory), the results are expected to be
    written to disk in wd, e.g. the "results" parameter is not used.
    :param wd: working directory of mpb calculation
    :param results: None
    :return: None
    """
    skip = ['epsilon', 'sym']
    # Read slices.
    results = []
    idx = 0
    while True:
        slice_path = os.path.join(wd, 'slice{}.json'.format(idx))
        if not os.path.isfile(slice_path):
            break
        with open(slice_path, 'r') as in_file:
            results.append(jsonpickle.decode(in_file.read()))
        idx += 1
    # Do the aggregation (rather ugly).
    aggregation = MpbResult(None, None, None, None, None, None, None)
    for key in results[0].__dict__:
        if results[0].__dict__[key] is None:
            aggregation.__dict__[key] = None
            continue
        elif key in skip:
            aggregation.__dict__[key] = results[0].__dict__[key]
            continue
        elif isinstance(results[0].__dict__[key], dict):
            aggregation.__dict__[key] = {subkey: np.concatenate(
                [result.__dict__[key][subkey] for result in results]) for subkey in results[0].__dict__[key]}
            continue
        aggregation.__dict__[key] = np.concatenate([result.__dict__[key] for result in results])
    # Dump the aggregated results.
    with open(os.path.join(wd, "mpb_result.json"), 'w+') as outfile:
        outfile.write(jsonpickle.encode(aggregation))
    # Remove result slices.
    for i in range(idx):
        os.remove(os.path.join(wd, "slice{}.json".format(i)))


def run_mpb_calculation(name, mpb_input, cores=None, wd=None, calc=None, agg=None, args=None, chunk_size=None):
    """
    Main entry point for all mpb calculations.
    :param name: name of the calculation, a folder with this name is created (and cleaned!) on execution
    :param mpb_input: inputs for the mpb calculation ("MpbInput")
    :param cores: number of cores to utilize, default behaviour is to pick the number of cores available on the system
    :param chunk_size: number of jobs dispatched per call, higher chunk_size reduces sync overhead/memory allocation
    :param wd: working directory for calculation (relative to the current working directory)
    :param calc: defines what to do with the mpb data, when they are available. Default behaviour is to dump them as
    a .json file, but other options include calculation of the DOS result on-the-fly. The latter option is particularly
    useful when the MPB data are too large to fit on the disk, e.g. when many k-points are evaluated. Current options 
    include "calc_dos_result", "calc_frac_dos_result" and "dump_mpb_result" (default)
    :param agg: defines how to aggregate the data. Default behaviour is no aggregation. Current options include 
    "agg_dos_result" and "agg_mpb_results". Default is no aggregation
    :param args: injection of additional arguments to the "calc" calculation, used e.g. for on-the-fly DOS calculation.
    See the selected calculation for details
    :return: None
    """
    # Check inputs.
    mpb_args = mpb_input.mpb_params
    sym = mpb_input.sym is not None and mpb_input.sym
    if not (mpb_args["resolution"] % 2) == 0 and sym:
        raise ValueError("Symmetry only supported for even resolutions.")
    if cores is None:
        cores = multiprocessing.cpu_count()
    calc = calc if calc is not None else dump_mpb_result
    chunk_size = 1 if chunk_size is None else chunk_size

    def build_info():
        mpb_info = mpb_input.__dict__.copy()
        mpb_info["mpb_grid"] = mpb_input.mpb_grid.__dict__.copy()
        info = {"name": name,
                "agg": False if agg is None else agg.__name__,
                "mpb_input": mpb_info,
                "calc_args": args}
        return info

    def build_slices(wd):
        slices = mpb_input.mpb_grid.get_slices()
        for i in range(len(slices)):
            slices[i]["name"] = name
            slices[i]["modes"] = mpb_input.modes
            slices[i]["wd"] = os.path.join(wd, "slice{}".format(i))
            slices[i].update(mpb_input.mpb_params)
            if args is not None:
                slices[i].update(args)  # Hack to inject additional arguments for the calculation itself
            if mpb_input.sym is not None:
                slices[i]["sym"] = mpb_input.sym
            if mpb_input.field is not None:
                slices[i]["field"] = mpb_input.field
        # Divide them into chunks.
        chunk_size_max = int(np.floor(len(slices) / min(cores, len(slices))))
        the_chunk_size = min(chunk_size, chunk_size_max)
        num_chunks = int(np.ceil(len(slices) / the_chunk_size))
        chunks = [slices[i * the_chunk_size:min((i + 1) * the_chunk_size, len(slices))] for i in range(num_chunks)]
        return {"slices": chunks}

    def build_result(wd, slice_map, entries):
        toc11 = time.time()
        if agg is not None:
            agg(wd, entries)
            # Log aggregation ended.
            toc12 = time.time()
            print("Aggregation took {:.2f}s.".format(toc12 - toc11))
        return None  # currently result is only written to file

    # Setup injections and dispatch.
    funcs = {"build_info": build_info, "build_slices": build_slices, "build_result": build_result, "calc": calc}
    dispatch(name, "mpb", funcs, wd, cores)


# endregion

# region Processing of mbp data

def dos_result_from_mpb_result(wd, w_bins, suffix=None, clean_up=False, record_ldos2d=False):
    """
    Calculation of DOS result from MPB result (this method is practically legacy, for most purposes the on-the-fly
    DOS calculation should be used, but the method remains for testing purposes).
    :param record_ldos2d: if True, LDOS2D is additionally calculated and saved
    :param suffix: add suffix to file name when saved
    :type wd: folder which contains the mpb result in slices (aggregated mpb result not supported)
    :param w_bins: bins for the DOS calculation
    :param clean_up: if True, all input files are deleted when the DOS calculation finishes
    :return: None
    """
    # Check inputs.
    name = os.path.split(wd)[-1]
    wd = os.path.dirname(wd)
    suffix = "_{}".format(suffix) if suffix is not None else ""

    def build_info():
        return {}

    def build_slices(wd):
        slices = []
        idx = 0
        while True:
            slice_path = os.path.join(wd, "slice{}.json".format(idx))
            if not os.path.isfile(slice_path):
                break
            slices.append(slice_path)
            idx += 1
        return {"slices": slices}

    def build_result(wd, slice_map, entries):
        file_name = "dos_result_dw{:.0e}{}.json".format(w_bins[1] - w_bins[0], suffix)
        with open(os.path.join(wd, file_name), 'w') as out_file:
            out_file.write(jsonpickle.encode(dos_result))
        return None

    def calc(slice_path):
        mpb_result = load(slice_path)
        populate_dos_result(dos_result, mpb_result, record_ldos2d)
        # Remove data if requested.
        if clean_up:
            os.remove(slice_path)

    mpb_result = load(os.path.join(wd, name, "slice0.json"))
    dos_result = allocate_dos_result(mpb_result, w_bins, record_ldos2d)
    # Setup injections and dispatch.
    funcs = {"build_info": build_info, "build_slices": build_slices, "build_result": build_result, "calc": calc}
    dispatch(name, "dos_from_mpb", funcs, wd, 1, False)  # multicore currently NOT supported


def calc_dos_result(args_list):
    """
    Calculation of DOS result. 
    :param args_list: list of MPB (see "calculate_mpb_result") and DOS calculation arguments which should contain 
    "dw", "w_max" and "record_ldos2d"
    :return: dictionary of DOS results
    """
    # Loop over slices in chunk.
    for i, args in enumerate(args_list):
        wd, mpb_result = calculate_mpb_result(args)
        # Read properties from slice.
        record_ldos2d = args["record_ldos2d"] if "record_ldos2d" in args.keys() else False
        # Allocate memory on first run.
        if i == 0:
            output = {}
        # Loop over binning resolutions.
        for dw in args["dw"]:
            key = "dw{:.0e}".format(dw)
            w_bins = np.arange(0, args["w_max"], dw)
            # Allocate dos object on first run.
            if key not in output:
                output[key] = allocate_dos_result(mpb_result, w_bins, record_ldos2d)
            # Fill in the data.
            populate_dos_result(output[key], mpb_result, record_ldos2d)
    return output


def agg_dos_results(wd, results):
    """
    Aggregation of DOS results. As DOS results are small (fit in memory), the results are expected to be
    available on the results object, i.e. no files are read thus reducing I/O.
    :param wd: working directory to which the aggregated DOS result is written
    :param results: DOS results to be aggregated
    :return: None
    """
    skip = ['dk', 'sym', 'w_bins', 'epsilon']
    # Do the aggregation.
    result = results[0]
    for key in result:
        dos_result = result[key].clone()
        for i in np.arange(1, len(results)):
            tmp = results[i][key]
            for subkey in dos_result.__dict__:
                if subkey in skip or (subkey is "ldos2d" and dos_result.__dict__[subkey] is None):
                    continue
                for subsubkey in dos_result.__dict__[subkey]:
                    dos_result.__dict__[subkey][subsubkey] += tmp.__dict__[subkey][subsubkey]
        # Dump the final result.
        file_name = "dos_result_{}.json".format(key)
        with open(os.path.join(wd, file_name), 'w') as out_file:
            out_file.write(jsonpickle.encode(dos_result))


def allocate_dos_result(mpb_result, w_bins, record_ldos2d=False):
    """
    Creates empty DOS result.
    :param mpb_result: mpb result object ("MpbResult")
    :param w_bins: vector of w bins
    :param record_ldos2d: if True ldos2d is recorded
    :return: empty DOS result object ("DosResult")
    """
    # Read properties from slice.
    modes = mpb_result.get_modes()
    space_res = len(mpb_result.epsilon)
    w_res = len(w_bins)
    dk = max(mpb_result.kx[1] - mpb_result.kx[0], mpb_result.ky[1] - mpb_result.ky[0])
    # Allocate memory.
    dos2d = {mode: np.zeros(w_res) for mode in modes}
    dos3d = {mode: np.zeros(w_res) for mode in modes}
    ldos2d = {mode: np.zeros((w_res, space_res)) for mode in modes} if record_ldos2d else None
    ldos3d = {mode: np.zeros((w_res, space_res)) for mode in modes}

    return DosResult(w_bins, dos2d, dos3d, ldos2d, ldos3d, mpb_result.epsilon, dk, mpb_result.sym)


def populate_dos_result(dos_result, mpb_result, record_ldos2d):
    """
    Fills empty DOS result with data from mpb result. 
    :param dos_result: dos result object ("DosResult")
    :param mpb_result: mpb result object ("MpbResult")
    :param record_ldos2d: if True ldos2d is recorded
    :return: None
    """
    # Read properties from slice.
    num_bands = mpb_result.get_num_bands()
    w_bins = dos_result.w_bins
    dw = w_bins[1] - w_bins[0]
    # Setup ky matrix with same as the ws (index mapping).
    ky_matrix = np.reshape(np.repeat(mpb_result.ky, num_bands), (len(mpb_result.ky), num_bands))
    # Loop over desired modes.
    for mode in mpb_result.get_modes():
        ws = mpb_result.ws[mode]
        fields = mpb_result.fields[mode]
        # Loop over bins.
        for i, w_bin in enumerate(w_bins):
            next_w_bin = w_bins[i + 1] if i < len(w_bins) - 1 else w_bin + dw
            # Determine relevant bins.
            in_bin = np.logical_and(w_bin <= ws, ws < next_w_bin)
            # If there are no entries, go to next.
            if not in_bin.any():
                continue
            # Calculate the dos items.
            dos2d, dos3d, ldos2d, ldos3d = eval_dos(ky_matrix, fields, mpb_result.epsilon, in_bin)
            # Append results.
            dos_result.dos2d[mode][i] += dos2d
            dos_result.dos3d[mode][i] += dos3d
            if record_ldos2d:
                dos_result.ldos2d[mode][i, :] += ldos2d
            dos_result.ldos3d[mode][i, :] += ldos3d


def eval_dos(ky_matrix, fields, epsilon, in_bin):
    """
    Calculation of DOS.
    :param ky_matrix: matrix of ky values
    :param fields: matrix of field vectors
    :param epsilon: vector of epsilon values
    :param in_bin: indices of values in current bin
    :return: calculated dos2d, dos3d, ldos2d, ldos3d
    """
    # Select entries in the bin.
    ky = ky_matrix[in_bin]
    abs_e_sq = fields[in_bin, :] / epsilon
    # Calculate DOS/LDOS.
    dos2d = np.sum(in_bin)
    dos3d = np.sum(ky * 2.0 * np.pi)
    ldos2d = np.sum(abs_e_sq, 0)
    ldos3d = np.sum((abs_e_sq.T * ky).T * 2.0 * np.pi, 0)
    return dos2d, dos3d, ldos2d, ldos3d


# endregion


# region Fractional LDOS

def calc_frac_dos_result(args_list):
    """
    Calculation of fractional DOS result on the fly. As DOS results are small (fit in memory), the results are expected
    to be available on the results object, i.e. no files are read thus reducing I/O. Should never be called directly,
    but only passed as a "calc" parameter to "run_mpb_calculation", typically with "agg_dos_results" as the "agg"
    parameter. The fractional calculation takes additional parameters "n_out", "theta_d" (detection cone half-angle,
    degrees) and "phi_d" (detection angles, degrees).
    :param args_list: list of mpb (see "calculate_mpb_result") and fractional DOS calculation arguments which should
    contain "dw", "w_max", "record_ldos2d", "theta_d", "phi_d", "n_out"
    :return: dictionary of DOS result objects
    """
    # Loop over slices in chunk.
    for i, args in enumerate(args_list):
        wd, mpb_result = calculate_mpb_result(args)
        # Read properties from slice.
        modes = mpb_result.get_modes()
        num_bands = mpb_result.get_num_bands()
        record_ldos2d = args["record_ldos2d"] if "record_ldos2d" in args.keys() else False
        theta_d = args["theta_d"]
        # Allocate memory on first run ONLY (performance critical).
        if i == 0:
            output = {}
        # Loop over binning resolutions.
        for dw in args["dw"]:
            w_bins = np.arange(0, args["w_max"], dw)
            dw = w_bins[1] - w_bins[0]
            # Loop over detection angle.
            for phi_d in args["phi_d"]:
                key = "dw{:.0e}_phi{}".format(dw, phi_d)
                # Allocate dos object on first run.
                if key not in output:
                    output[key] = allocate_dos_result(mpb_result, w_bins, record_ldos2d)
                # Setup ky matrix with same as the bands (index mapping).
                ky_matrix = np.reshape(np.repeat(mpb_result.ky, num_bands), (len(mpb_result.ky), num_bands))
                # Loop through desired modes.
                for mode in modes:
                    ws = mpb_result.ws[mode]
                    fields = mpb_result.fields[mode]
                    # Calculate which modes that escapes the structure.
                    escaped = ky_matrix <= ws * args["n_out"]
                    if not escaped.any():
                        continue
                    # Calculate which modes that are collected at the angle phi_k.
                    phi_k = np.empty_like(ky_matrix)
                    phi_k.fill(180)  # dummy value, indicates that the mode is NOT collected
                    phi_k[escaped] = np.rad2deg(np.arcsin(ky_matrix[escaped] / (ws[escaped] * args["n_out"])))
                    collected = np.logical_and(phi_d - theta_d <= phi_k, phi_k <= phi_d + theta_d)
                    if not collected.any():
                        continue
                    # Calculate circle arc ratios for weighting.
                    fd_matrix = np.zeros_like(ky_matrix)
                    fd_matrix[collected] = frac_detectable(theta_d, phi_d, ws[collected], ky_matrix[collected])
                    # Calculate dos and ldos.
                    for j, w_bin in enumerate(w_bins):
                        next_w_bin = w_bins[j + 1] if j < len(w_bins) - 1 else w_bin + dw
                        # Determine relevant bins.
                        in_bin = np.logical_and(w_bin <= ws, ws < next_w_bin)
                        # Aggregate condition.
                        valid = np.logical_and(collected, in_bin)
                        # Calculate DOS/LDOS.
                        wt_matrix = ky_matrix * fd_matrix
                        dos2d, dos3d, ldos2d, ldos3d = eval_dos(wt_matrix, fields, mpb_result.epsilon, valid)
                        # Append results.
                        output[key].dos2d[mode][j] += dos2d
                        output[key].dos3d[mode][j] += dos3d
                        if record_ldos2d:
                            output[key].ldos2d[mode][j, :] += ldos2d
                        output[key].ldos3d[mode][j, :] += ldos3d
    return output

# endregion
