from scipy.signal import savgol_filter as sg
from scipy.signal import argrelextrema
from scipy.ndimage.filters import gaussian_filter1d as gf1d
from dos_analytic import *


def assemble_dos_data(dos_result, key, modes=None, rel=True, flip=True, fltr=None):
    """
    Method for assembling a dos vector/matrix from a dos result object ("DosResult").
    :param dos_result: dos result object ("DosResult")
    :param key: which dos type, can be "dos2d", "dos3d", "ldos2d" or "ldos3d"
    :param modes: list of modes to include, default is both te and tm
    :param rel: calculate the relative (rather than absolute) dos
    :param flip: do spatial flipping of ldos so as to get low index layer first, then high index layer
    :param fltr: filter do apply to dos
    :return: assembled dos vector/matrix
    """
    modes = modes if modes is not None else ["te", "tm"]
    data = np.zeros_like(dos_result.__dict__[key][modes[0]])
    # Add modes, typically we want both tm and te.
    for mode in modes:
        data += dos_result.__dict__[key][mode]
    # Calculate the relative DOS/LDOS.
    if rel:
        if "3d" not in key:
            raise ValueError("Relative values only supported for DOS3D/LDOS3D so far.")
        eps_low = np.min(dos_result.epsilon)
        dos3d = dos3d_analytic(np.sqrt(eps_low), dos_result.dk, dos_result.w_bins)
        data = (data.T / dos3d).T * (eps_low * len(dos_result.epsilon) if "ldos" in key else 1)
    # Flipping of LDOS for visualization purposes.
    if "ldos" in key and flip:
        data = rearrange_dos_data(data, dos_result.epsilon)
    # Filtering functionality, used to filter out binning noise.
    if fltr is not None:
        data = filter_ldos(data, fltr) if "ldos" in key else fltr(data)

    return data


def rearrange_dos_data(data, epsilon):
    """
    Rearrangement in space of ldos data so that the low index layer points occur first, then the high index points.
    :param data: data to rearrange
    :param epsilon: epsilon vector
    :return: rearranged data
    """
    high_part = data[:, abs(epsilon - min(epsilon)) > abs(epsilon - max(epsilon))]
    low_part = data[:, abs(epsilon - min(epsilon)) < abs(epsilon - max(epsilon))]
    low_center = int(np.ceil(low_part.shape[1] / 2))
    low_part = np.concatenate((low_part[:, low_center:], low_part[:, :low_center]), 1)
    return np.concatenate((low_part, high_part), 1)


# region Filtering

def filter_ldos(ldos, fltr):
    """
    Apply some filter to ldos matrix across the omega dimension.
    :param ldos: the ldos matrix
    :param fltr: the filter to apply
    :return: filtered ldos
    """
    for i in range(ldos.shape[1]):
        ldos[:, i] = fltr(ldos[:, i])
    return ldos


def default_filter(x, sigma=None):
    """
    Default filter option.
    :param x: vector to filter
    :param sigma: sigma for the gaussian filter
    :return: filtered vector
    """
    return gf1d(x, 1 if sigma is None else sigma)


def max_min_filter(x, sg1=3, sg2=31):
    """
    Filter than uses maximum/minimum point to do a piece-wise Savitzky-Golay filtering. Will typically require manuel
    tuning, but gives very nice results.
    :param x: vector to filter
    :param sg1: initial filtering window size
    :param sg2: final filtering window size
    :return: filtered vector
    """
    # First, do pre filtering to remove very high frequent noise.
    result = sg(x, sg1, 1) if sg1 is not None else x
    # Next, locate maxima/minima that we want to keep.
    idx_max = argrelextrema(x, np.greater, order=500)
    idx_min = argrelextrema(-1.0 * x, np.greater, order=500)
    all_idx = sorted(np.concatenate((idx_max[0], idx_min[0]), 0))
    # Remove maxima/minima is thee are too close to start/end.
    if all_idx[0] < sg2:
        del all_idx[0]
    if (len(result) - all_idx[-1]) < sg2:
        del all_idx[-1]
    # Then, do piece-wise filtering.
    offset = 0
    for i in range(len(all_idx)):
        next_idx = all_idx[i]
        result[offset:next_idx] = sg(result[offset:next_idx], sg2, 2)
    result[offset:] = sg(result[offset:], sg2, 2)
    return result

# endregion
