import matplotlib.pyplot as plt

from dos_core import *
from rem_core import *


def band_gap_edges(n_high, num_bands):
    """
    Get the omega values of the edges of the band gap(s).
    :param n_high: refractive index of high index materials
    :return: vector of band edge points
    """
    wd = "../tmp_test"
    step = 1e-3  # if step is None else step
    # Calculate band data.
    key = "band_calc"
    grid = MpbGrid1D(0.5 - 3 * step, 0.5, step, 'x', 0 + step / 2.0)  # look at kx at ky = 0 where the gap occurs
    # Setup calculation arguments.
    mpb_params = {"n_low": 1.5, "n_high": n_high, "num_bands": num_bands, "resolution": 32}
    # Run calculation.
    mpb_input = MpbInput(mpb_params, grid, field=False)
    run_mpb_calculation(key, mpb_input, agg=agg_mpb_results, wd=wd)
    result = load(os.path.join(wd, key, "mpb_result.json"))
    # Derive band gaps.
    return np.array(result.ws['te'][0])


def find_omega(lmbd, n_low, n_high):
    """
    Determination of maximum omega value needed to cover all wanted emissions and designs.
    :param lmbd: maximum design wavelength considered
    :param n_low: refractive index of low index material
    :param n_high: refractive index of high index material
    :return: minimum omega, maximum omega needed to cover all wanted emissions and designs
    """
    lmb_min = np.min(np.array(hofmann2016()['lmbs'])[np.array(hofmann2016()['lmbs']) != 0])
    lmb_max = np.max(np.array(hofmann2016()['lmbs']))
    w_min = (n_low + n_high) / (4 * n_low * n_high) * lmbd / lmb_max
    w_max = (n_low + n_high) / (4 * n_low * n_high) * lmbd / lmb_min
    return w_min, w_max


def find_bands_and_ky_max(n_low, n_high, scale=1.0, omega_max=1.2, visual=False):
    """
    Calculation of ky_max and the number of bands needed for a given material combination (n_low, n_high) and omega_max.
    :param n_low: refractive index of low index material
    :param n_high: refractive index of high index material
    :param scale: scaling of omega (typically not used)
    :param omega_max: maximum omega value needed to cover all wanted emissions and designs
    :param visual: if True, results are plotted
    :return: None
    """
    omega_max *= scale  # Maximum frequency, step i k
    # Do the calculations.
    cwd = prepare_mpb_calculation(n_low, n_high, "..")
    # Find the maximum ky values.
    result = load("{}/mpb_result.json".format(cwd))
    num_bands = result.ws["te"].shape[1]
    # Log findings.
    print("")
    print("***** Determine bands/ky_max START ***** ")
    # Figure out what we need.
    max_ky = 0
    relevant_bands = []
    for i in range(num_bands):
        relevant_te = result.ws["te"][:, i] <= omega_max
        relevant_tm = result.ws["tm"][:, i] <= omega_max
        relevant = np.logical_or(relevant_te, relevant_tm)
        if not relevant.any():
            continue
        relevant_bands.append(i)
        max_ky = max(max_ky, np.max(result.ky[relevant]))
    print("Maximum ky is: {}".format(max_ky))
    print("Relevant bands are: {}".format(relevant_bands))
    print("***** Determine bands/ky_max END ***** ")
    if not visual:
        return len(relevant_bands), max_ky
    # Plot stuff for verification.
    plt.figure(figsize=(20, 10))
    plt.subplot(1, 2, 1)
    plt.plot([0, 10], [omega_max, omega_max], color='grey')
    plt.plot([max_ky, max_ky], [0, 10], color='grey')
    plt.xlim(0, max_ky + 2)
    plt.ylim(0, omega_max + 2)
    plt.subplot(1, 2, 2)
    plt.plot([0, 10], [omega_max, omega_max], color='grey')
    plt.plot([max_ky, max_ky], [0, 10], color='grey')
    plt.xlim(0, max_ky + 2)
    plt.ylim(0, omega_max + 2)
    # Iterative bands.
    for i in range(num_bands):
        omega_te = result.ws["te"][:, i]
        omega_tm = result.ws["tm"][:, i]
        kx_min = result.kx == np.min(result.kx)
        kx_max = result.kx == np.max(result.kx)
        plt.subplot(1, 2, 1)
        plt.title("kx = {}".format(np.min(result.kx)))
        if i in relevant_bands:
            plt.plot(result.ky[kx_min], omega_te[kx_min], label="te{}".format(i))
            plt.plot(result.ky[kx_min], omega_tm[kx_min], "--", label="tm{}".format(i))
        else:
            plt.plot(result.ky[kx_min], omega_te[kx_min], color='pink')
            plt.plot(result.ky[kx_min], omega_tm[kx_min], color='pink')
        plt.legend(loc=4, ncol=4)
        plt.subplot(1, 2, 2)
        plt.title("kx = {}".format(np.max(result.kx)))
        if i in relevant_bands:
            plt.plot(result.ky[kx_max], omega_te[kx_max], label="te{}".format(i))
            plt.plot(result.ky[kx_max], omega_tm[kx_max], "--", label="tm{}".format(i))
        else:
            plt.plot(result.ky[kx_max], omega_te[kx_max], color='pink')
            plt.plot(result.ky[kx_max], omega_tm[kx_max], color='pink')
        plt.legend(loc=4, ncol=4)
    plt.show()


def prepare_mpb_calculation(n_low, n_high, wd):
    """
    Setup a quick mpb calculation for band investigation.
    :param n_low: refractive index of low index material
    :param n_high: refractive index of high index material
    :param wd: working directory
    :return: path to folder containing mpb results
    """
    wd = "../tmp_test"
    key = "prepare_n_low_{}_n_high_{}".format(n_low, n_high)
    # Note: The maximum possible number of bands, assumed to be 10, is included in the trail simulation.
    mpb_params = {"n_low": n_low, "n_high": n_high, "num_bands": 10, "resolution": 32}
    # Note: Grid is designed with ONLY start/end points in x (where band energy is lowest/highest).
    grids = {"kxlow": MpbGrid1D(0, 10, 1e-2, 'y', 1e-2 / 2.0), "kxhigh": MpbGrid1D(0, 10, 1e-2, 'y', 0.5 - 1e-2 / 2.0)}
    for entry in grids:
        mpb_input = MpbInput(mpb_params, grids[entry])
        mpb_input.field = None  # Disable field evaluation (not necessary for test calculations).
        run_mpb_calculation("{}_{}".format(key, entry), mpb_input, cores=1, agg=agg_mpb_results, wd=wd)
    # Hack: Merge the results.
    if os.path.exists(os.path.join(wd, key)):
        shutil.rmtree(os.path.join(wd, key))
    os.makedirs(os.path.join(wd, key))
    os.rename(os.path.join(wd, "{}_{}".format(key, "kxlow"), "mpb_result.json"), os.path.join(wd, key, "slice0.json"))
    os.rename(os.path.join(wd, "{}_{}".format(key, "kxhigh"), "mpb_result.json"), os.path.join(wd, key, "slice1.json"))
    agg_mpb_results(os.path.join(wd, key), None)
    return os.path.join(wd, key)
