from bragg_objects import *
from bragg_utils import *
from tmmpy_utils import *


def run_irr_calculation(name, bs, param_map, res=None, wd=None, cores=None, calc=None):
    """
    Main entry point of irradiance calculation.
    :param name: name of the calculation, a folder with this name is created (and cleaned!) on execution
    :param bs: bragg stack object ("BraggStack")
    :param param_map: map of parameters, should contain design wavelengths "lmbd", excitation wavelengths "lmb" and
    incident angles "ang"
    :param res: number of points evaluated per active layer
    :param wd: working directory in which folder with result files is created
    :param cores: number of cores to utilize, default behaviour is to pick the number of cores available on the system
    :param calc: calculation function, default is "calc_irr_result"
    :return: irradiance result object ("IrrResult")
    """
    # Initialize default parameters.
    res = 16 if res is None else res  # minimum recommended resolution, for publications try higher
    calc = calc_irr_data if calc is None else calc
    cores = cores if cores is not None else multiprocessing.cpu_count()

    def build_info():
        info = {"name": name, "res": res, "bs": jsonpickle.encode(bs), "params": {}}
        for key in param_map:
            value = param_map[key]
            info["params"][key] = value.tolist() if isinstance(value, (np.ndarray, np.generic)) else value
        return info

    def build_slices(cwd):
        param_names = ["lmbd", "lmb", "ang"]
        param_values = [param_map[key].tolist() for key in param_names]
        slices = []
        for lmbd in param_map["lmbd"]:
            for lmb in param_map["lmb"]:
                for ang in param_map["ang"]:
                    slices.append({"lmbd": lmbd, "lmb": lmb, "ang": ang, "res": res, "bs": bs})
        return {"slices": slices, "param_names": param_names, "param_values": param_values}

    def build_result(cwd, slice_map, entries):
        irr_result = IrrResult(slice_map["param_names"], slice_map["param_values"], {'num_al': bs.num_al}, entries)
        with open(os.path.join(cwd, "irr_result.json"), 'w') as out_file:
            out_file.write(jsonpickle.encode(irr_result))
        return irr_result

    # Setup injections and dispatch.
    funcs = {"build_info": build_info, "build_slices": build_slices, "build_result": build_result, "calc": calc}
    return dispatch(name, "IRR", funcs, wd, cores)


def calc_irr_data(args):
    """
    Calculation of irradiance data.
    :param args: irradiance calculation parameters (see "run_irr_calculation")
    :return: vector of relative irradiance in active layers
    """
    args["bs"].lmbd = args["lmbd"]
    pts, irr_brg, irr_ref = calc_irr(args["lmb"], args["ang"], args["bs"], args["res"])
    return irr_brg, irr_ref


def default_irr_params(lmbds, angs=None, lmbs=None):
    """
    Construction of default parameter set for irradiance calculation.
    :param lmbds: list of design wavelengths
    :param angs: list of incidence angles
    :param lmbs: list of exciation wavelengths
    :return: dictionary of parameters for irradiance calculation
    """
    return {"lmbd": np.array(lmbds),
            "ang": np.array([0]) if angs is None else np.array(angs),
            "lmb": np.array([1523]) if lmbs is None else np.array(lmbs)}
