import shutil
import scipy.integrate
import scipy.interpolate

from scipy.constants import c, hbar, epsilon_0, pi  # , e, m_e
from bragg_objects import *
from bragg_utils import *


# region REM parameter sets

def hofmann2016_original():
    """
    Original physical parameters used in Hofmann 2016 optics express paper.
    :return: dictionary of physical parameters
    """
    args = hofmann2016()
    args["k_etu"] = [6.0417e19, 3.3431e17, 8.4584e18, 3.5445e22, 2.0139e18, 1.6917e19, 3.5848e19, 2.0139e20, 3.0209e22]
    args["k_cr"] = [4.8335e17, 2.6181e19, 7.2501e18, 3.4237e14, 8.8612e20, 7.6529e24, 4.0278e24, 3.7459e13, 1.3292e23]
    args["k12"] = 1.69800e-14
    args["k21"] = 9.30000e-15
    args["k24"] = 1.12936e-11
    args["k46"] = 3.55470e-15
    args["epsilon_0"] = 8.954187e-12
    args["lmbs"] = [[0, 1558, 984, 814, 660, 536],
                    [0, 0, 2672, 1705, 1146, 817],
                    [0, 0, 0, 4711, 2008, 1177],
                    [0, 0, 0, 0, 3499, 1568],
                    [0, 0, 0, 0, 0, 2842],
                    [0, 0, 0, 0, 0, 0]]
    return args


def hofmann2016():
    """
    The parameters stated here are almost equal to Hofmann2016. The only exceptions are k_etu/k_cr taken from Fischer
    thesis page 104, which differ slightly. As far as i (emher) have observed, the model output is not affected.
    :return: dictionary of physical parameters
    """
    args = {"epsilon_0": epsilon_0}
    # Material parameters.
    args["er_doping"] = 0.25
    args["refractive_index"] = 1.5
    # Matrix pre factor (Fischer 2011 eq 10).
    args["mpr_pre"] = 1.0e8
    # Parameters describing the transition probabilities, Fischer thesis page 104.
    args["k12"] = 1.7e-14  # Parameter for GSA
    args["k21"] = 7.1e-15  # Parameter for stimulated emission from 2 -> 1
    args["k24"] = 1.3e-11  # Parameter for ESA1
    args["k46"] = 5.7e-15  # Parameter for ESA2
    # Overlap integrals for the energy transfer in [1/eV], Fisher thesis table on page 105.
    # 2,2->1,4, 2,4->1,6, 4,4->2,6, 3,3->1,6, 4,2->1,6, 5,3->2,6, 3,5->2,6, 3,2->1,5, 4,3->2,5
    args["k_etu"] = [6.1e19, 3.3e17, 8.6e18, 3.6e22, 2.0e18, 1.7e19, 3.6e19, 2.01e20, 3.0e22]
    args["k_cr"] = [5.0e17, 2.6e19, 7.2e18, 3.5e14, 9.0e20, 7.7e24, 4.2e24, 3.7e13, 1.4e23]
    # Crystal material constant [1/J]. Calculated based on data (from book by Garcia Sole) as
    # (ln(MPR(0)-ln(RateMPR-Buch))/(hbar*Omega), energy distance as for LaCl.
    args["kappa"] = 2.15e20
    # New Values: Diss Stefan, page 231, table A.1, diagonal elements set to zero (makes sense physically, i guess...).
    args["dipole_square"] = (np.array([[0, 7.19, 2.04, 0.12, 2.56, 1.10],
                                       [7.19, 0, 5.59, 3.50, 0.53, 1.68],
                                       [2.04, 5.59, 0, 0.79, 6.50, 0.38],
                                       [0.12, 3.50, 0.79, 0, 0.54, 1.31],
                                       [2.56, 0.53, 6.50, 0.54, 0, 0.11],
                                       [1.10, 1.68, 0.38, 1.31, 0.11, 0]]) * 1e-62).tolist()
    # New values: Emission frequencies (from R-Simulation by Benni, absorption frequency not stated, set to zero).
    args["freqs"] = [[0, 1209093019934110, 1913979982210212, 2313825394451770, 2852179102029885, 3514888986659564],
                     [0, 0, 704886962276102, 1104732374517659, 1643086082095775, 2305795966725453],
                     [0, 0, 0, 399845412241558, 938199119819673, 1600909004449352],
                     [0, 0, 0, 0, 538353707578116, 1201063592207794],
                     [0, 0, 0, 0, 0, 662709884629679],
                     [0, 0, 0, 0, 0, 0]]
    # Wavelength of the different transitions.
    with np.errstate(divide='ignore', invalid='ignore'):
        args["lmbs"] = c / np.array(args["freqs"]) * 1e9 * pi * 2
        args["lmbs"][np.isinf(args["lmbs"])] = 0
        args["lmbs"] = args["lmbs"].tolist()
    return args


# Note: Add more parameter sets here.


# endregion

# region Parameter initialization

def initialize_rem_matrices(args):
    """
    Various physical parameters relevant for the REM are initialized here. The setup is equal to Hofmann2016 except:
    * the STE elements are divided by the correct g-factor (for level 2 rather than level 1)
    * epsilon is the correct value (it's off by around a percent in Hofmann 2016)
    * the k_21 degeneracy correction is moved to the parameter definition (not included in Hofmann 2016)
    As far as i (emher) have observed, the model output is not affected by the corrections.
    :param args: physical parameter set (see e.g. "hofmann2016")
    :return: matrices for rem: spe, gsa, esa, ste, mpr, k_etu, mat_etu, k_cr, mat_cr, dp_sq, einstein
    """

    # region Setup parameters

    # Distance [m] between ions (depends on doping, which is assumed to be 25%).
    radius = ((5.9757e-10 * 5.9757e-10 * 3.5305e-10 * np.sqrt(3.0) / 2.0) / (1.5 * args["er_doping"])) ** (1.0 / 3.0)
    # Correction factor(E_loc/E)^2 from Henderson page 157, eq 4.61 (we look ONLY at the low index region!).
    ri = args["refractive_index"]
    e_loc_corr = ((ri ** 2.0 + 2.0) / 3.0) ** 2
    # Matrix pre factors, Hofmann thesis : einstein eq 2.28, gsa eq 5.12, et eq 2.33.
    mpr_pre = args["mpr_pre"]
    gsa_pre = pi / (3.0 * ri ** 2.0 * hbar ** 2.0 * args["epsilon_0"]) * e_loc_corr * (ri / c) * args["irr"]
    et_pre = (12.0 * hbar * pi * args["epsilon_0"] ** 2.0 * ri ** 4.0 * radius ** 6.0) ** -1.0
    einstein_pre = ri * e_loc_corr / (3.0 * pi * args["epsilon_0"] * hbar * c ** 3.0)
    # Degeneracy of the energy levels, i.e. 2(J+1) except for level 6 (+ 12 from level 7), see fig 1 in Fischer2011.
    g = np.array([16.0, 14.0, 12.0, 10.0, 10.0, 16.0])
    # Dipole matrix elements squared.
    dp_sq = np.array(args["dipole_square"])
    # Emission frequencies.
    freq = np.array(args["freqs"])

    # endregion

    # region Calculate matrices

    # Overlap integrals for the energy transfer upconversion (ETU) and cross relaxation (CR).
    k_etu, k_cr = np.array(args["k_etu"]) * et_pre, np.array(args["k_cr"]) * et_pre
    # Matrix for ETU.
    mat_etu = np.empty((len(g), len(k_etu)))
    mat_etu[:, 0] = np.array([+1.0, -2.0, +0.0, +1.0, +0.0, +0.0]) * dp_sq[1, 0] * dp_sq[1, 3]
    mat_etu[:, 1] = np.array([+1.0, -1.0, +0.0, -1.0, +0.0, +1.0]) * dp_sq[1, 0] * dp_sq[3, 5]
    mat_etu[:, 2] = np.array([+0.0, +1.0, +0.0, -2.0, +0.0, +1.0]) * dp_sq[3, 1] * dp_sq[3, 5]
    mat_etu[:, 3] = np.array([+1.0, +0.0, -2.0, +0.0, +0.0, +1.0]) * dp_sq[2, 0] * dp_sq[2, 5]
    mat_etu[:, 4] = np.array([+1.0, -1.0, +0.0, -1.0, +0.0, +1.0]) * dp_sq[3, 0] * dp_sq[1, 5]
    mat_etu[:, 5] = np.array([+0.0, +1.0, -1.0, +0.0, -1.0, +1.0]) * dp_sq[4, 1] * dp_sq[2, 5]
    mat_etu[:, 6] = np.array([+0.0, +1.0, -1.0, +0.0, -1.0, +1.0]) * dp_sq[2, 1] * dp_sq[4, 5]
    mat_etu[:, 7] = np.array([+1.0, -1.0, -1.0, +0.0, +1.0, +0.0]) * dp_sq[2, 0] * dp_sq[1, 4]
    mat_etu[:, 8] = np.array([+0.0, +1.0, -1.0, -1.0, +1.0, +0.0]) * dp_sq[3, 1] * dp_sq[2, 4]
    # Matrix for CR.
    mat_cr = np.empty((len(g), len(k_cr)))
    mat_cr[:, 0] = np.array([-1.0, +2.0, +0.0, -1.0, +0.0, +0.0]) * dp_sq[3, 1] * dp_sq[0, 1]
    mat_cr[:, 1] = np.array([-1.0, +1.0, +0.0, +1.0, +0.0, -1.0]) * dp_sq[5, 3] * dp_sq[0, 1]
    mat_cr[:, 2] = np.array([+0.0, -1.0, +0.0, +2.0, +0.0, -1.0]) * dp_sq[5, 3] * dp_sq[1, 3]
    mat_cr[:, 3] = np.array([-1.0, +0.0, +2.0, +0.0, +0.0, -1.0]) * dp_sq[5, 2] * dp_sq[0, 2]
    mat_cr[:, 4] = np.array([-1.0, +1.0, +0.0, +1.0, +0.0, -1.0]) * dp_sq[5, 1] * dp_sq[0, 3]
    mat_cr[:, 5] = np.array([+0.0, -1.0, +1.0, +0.0, +1.0, -1.0]) * dp_sq[5, 2] * dp_sq[1, 4]
    mat_cr[:, 6] = np.array([+0.0, -1.0, +1.0, +0.0, +1.0, -1.0]) * dp_sq[5, 4] * dp_sq[1, 2]
    mat_cr[:, 7] = np.array([-1.0, +1.0, +1.0, +0.0, -1.0, +0.0]) * dp_sq[4, 1] * dp_sq[0, 2]
    mat_cr[:, 8] = np.array([+0.0, -1.0, +1.0, +1.0, -1.0, +0.0]) * dp_sq[4, 2] * dp_sq[1, 3]
    # Calculate einstein coefficients for spontaneous emission A_ij [1/s], Fischer 2011 eq 23.
    einstein = np.zeros((6, 6))
    for i in range(5):
        for j in range(i + 1, 6):
            einstein[i, j] = dp_sq[i, j] * freq[i, j] ** 3.0 / g[j]
    einstein *= einstein_pre
    # Spontaneous emission (SE), fischer2011 eq 4.
    spe = spe_from_einstein(einstein)
    # Ground/excited state absorption (GSA/ESA) + stimulated emission (STE), Hofmann thesis eq 5.11 and 5.13.
    gsa = np.zeros((6, 6))
    gsa[0, 0] = -dp_sq[0, 1] * args["k12"] / g[0]
    gsa[1, 0] = dp_sq[0, 1] * args["k12"] / g[0]
    gsa *= gsa_pre * args["abs_er"]
    esa = np.zeros((6, 6))
    esa[1, 1] = -dp_sq[1, 3] * args["k24"] / g[1]
    esa[3, 1] = dp_sq[1, 3] * args["k24"] / g[1]
    esa[3, 3] = -dp_sq[3, 5] * args["k46"] / g[3]
    esa[5, 3] = dp_sq[3, 5] * args["k46"] / g[3]
    esa *= gsa_pre
    ste = np.zeros((6, 6))
    ste[0, 1] = dp_sq[0, 1] * args["k21"] / g[1]  # NOTE: Division by g[1] corrected from Hofmann code
    ste[1, 1] = -dp_sq[0, 1] * args["k21"] / g[1]  # NOTE: Division by g[1] corrected from Hofmann code
    ste *= gsa_pre
    # Multi phonon relaxation, Hofmann thesis eq 5.14.
    mpr = np.zeros((6, 6))
    for i in range(0, 5):
        mpr[i + 0, i + 1] = +exp(-args["kappa"] * hbar * freq[i, i + 1])
        mpr[i + 1, i + 1] = -exp(-args["kappa"] * hbar * freq[i, i + 1])
    mpr *= mpr_pre

    # endregion

    return spe, gsa, esa, ste, mpr, k_etu, mat_etu, k_cr, mat_cr, dp_sq, einstein


def spe_from_einstein(einstein):
    """
    Calculation of spe matrix from Einstein coefficients. 
    :param einstein: Einstein coefficients
    :return: spe matrix
    """
    spe = np.zeros((6, 6))
    for j in range(1, 6):
        minus_jj_spe = 0
        for i in range(0, j):
            spe[i, j] = einstein[i, j]
            minus_jj_spe = minus_jj_spe - einstein[i, j]
        spe[j, j] = minus_jj_spe
    return spe


def ldos_correction(lmbs, ldos_input, lmbd, res):
    """
    Calculate einstein coefficient correction matrices (due to change of ldos). The result is a list of matrices with
    length equal to res, i.e., each entry corresponds to the corrections at a spatial point in the low index region.
    :param lmbs: emission wavelength matrix
    :param ldos_input: rem dos input object ("RemDosInput") 
    :param lmbd: design wavelength
    :param res: spatial resolution of the Wigner Seitz unit cell
    :return: list of correction matrices for Einstein coefficients
    """
    n_low, n_high = ldos_input.n_low, ldos_input.n_high
    # NOTE: This method is NOT very efficient, could be speed up be a pre processing step (but probably not necessary)
    corr_matrices = []
    x_interp = list(np.linspace(0, 1, res, endpoint=True))
    # Pre calculate LDOS correction factors (important for performance, 25X gain in speed of LDOS pre processing).
    result_map = {}
    unique_lmbs = list(set(lmbs.flatten()))
    for lmb in unique_lmbs:
        if lmb == 0:
            continue
        w = (n_low + n_high) / (4 * n_low * n_high) * lmbd / lmb
        ldos_space = ldos_input.ldos_rel[(np.abs(ldos_input.w_bins - w)).argmin(), :]
        interp = scipy.interpolate.interp1d(np.linspace(0, 1, len(ldos_space), endpoint=True), ldos_space)
        result_map[lmb] = interp(x_interp)
    # Fill values into the matrices (could maybe be done in a smarter way).
    for i in range(res):
        matrix = np.zeros_like(lmbs, dtype=float)
        for j in range(lmbs.shape[0]):
            for k in range(lmbs.shape[1]):
                if lmbs[j, k] == 0:
                    continue
                matrix[j, k] = result_map[lmbs[j, k]][i]
        corr_matrices.append(matrix)
    return corr_matrices


# endregion

# region REM calculation setup

def run_rem(name, rem_input, slice_gen=None, res=None, wd=None, atol=None, cores=None):
    """
    Main entry point for rate equation model simulations.
    :param name: output files are stored in a new folder with this name
    :param rem_input: rem input object ("RemInput")
    :param slice_gen: slice generator, default is "default_slice_gen"
    :param res: spatial resolution of the Wigner Seitz unit cell
    :param wd: working directory in which the output is created
    :param atol: absolute tolerance of inner solver
    :param cores: number of cores used for simulation
    :return: rem result object ("RemResult")
    """
    # Initialize default parameters.
    res = 8 if res is None else res  # minimum recommended resolution, for publications try higher
    atol = 1e-6 if atol is None else atol  # maximum recommended atol, for publications try lower
    cores = cores if cores is not None else multiprocessing.cpu_count()
    slice_gen = default_slice_gen if slice_gen is None else slice_gen
    calc = calculate_rem_data  # will this ever change?

    def build_info():
        info = {"name": name, "res": res, "atol": atol, "slice_gen": slice_gen.__name__, "params": {},
                "irr_input": rem_input.irr_input.get_info() if rem_input.irr_input is not None else None}
        for key in rem_input.param_map:
            value = rem_input.param_map[key]
            if hasattr(value[0], "get_info"):
                value = [item.get_info() for item in value]
            info["params"][key] = value.tolist() if isinstance(value, (np.ndarray, np.generic)) else value
        return info

    def build_slices(cwd):
        param_names, param_values, slices = slice_gen(rem_input, res)
        for slice in slices:
            slice["atol"] = atol
        return {"slices": slices, "param_names": param_names, "param_values": param_values}

    def build_result(cwd, slice_map, entries):
        rem_result = RemResult(slice_map["param_names"], slice_map["param_values"], entries)
        with open(os.path.join(cwd, "rem_result.json"), 'w') as out_file:
            out_file.write(jsonpickle.encode(rem_result))
        return rem_result

    # Setup injections and dispatch.
    funcs = {"build_info": build_info, "build_slices": build_slices, "build_result": build_result, "calc": calc}
    return dispatch(name, "REM", funcs, wd=wd, cores=cores)


def calculate_rem_data(args):
    """
    Calculation of rem data.
    :param args: rem calculation parameters (see "run_rem")
    :return: list of rem data
    """
    spe, gsa, esa, ste, mpr, k_etu, mat_etu, k_cr, mat_cr, dp_sq, einstein = initialize_rem_matrices(args["params"])
    # 1) Solve rate equation model for the reference system.
    irr_ref = args["irr_ref"]
    gsa_ref, esa_ref, ste_ref = gsa * irr_ref, esa * irr_ref, ste * irr_ref
    ts, xs = solve_uc_ode(spe, gsa_ref, esa_ref, ste_ref, mpr, k_etu, mat_etu, k_cr, mat_cr, args["atol"])
    lum_ref, abs_ref = calc_lum_abs(xs[-1, :], einstein, gsa_ref, esa_ref)
    vec_ref = xs[-1, :]
    # 2) Loop over spatial coordinates in the bragg stack.
    res = args["res"]
    num_al = int(len(args["irr_brg"]) / res)
    vec_brg, lum_brg, abs_brg = np.empty((num_al * res, 6)), np.empty((num_al * res, 6)), np.empty(num_al * res)
    for i in range(num_al):
        for j in range(res):
            # Modify irradiance (from TMM calculation).
            irr_brg = args["irr_brg"][i * res + j]
            # Modify einstein coefficients (from LDOS calculation) and construct new SPE matrix.
            einstein_brg = einstein * args["ldos_correction"][j]
            se_brg = spe_from_einstein(einstein_brg)
            # Solve rate equation model for for the bragg stack.
            gsa_brg, esa_brg, ste_brg = gsa * irr_brg, esa * irr_brg, ste * irr_brg
            ts, xs = solve_uc_ode(se_brg, gsa_brg, esa_brg, ste_brg, mpr, k_etu, mat_etu, k_cr, mat_cr, args["atol"])
            lum_tmp, abs_tmp = calc_lum_abs(xs[-1, :], einstein_brg, gsa_brg, esa_brg)
            # Collect results.
            vec_brg[i * res + j, :], lum_brg[i * res + j, :], abs_brg[i * res + j] = xs[-1, :], lum_tmp, abs_tmp
    d_tot = num_al * args["lmbd"] / (4 * args["params"]["refractive_index"])
    return [vec_ref, lum_ref * d_tot, abs_ref * d_tot,
            np.mean(vec_brg, 0), np.mean(lum_brg, 0) * d_tot, np.mean(abs_brg) * d_tot]


def solve_uc_ode(spe, gsa, esa, ste, mpr, k_etu, mat_etu, k_cr, mat_cr, atol):
    """
    Given the upconversion parameters, solve the differential equation (emulate ode15s) and return the result.
    :param spe: spontaneous emission matrix
    :param gsa: ground state absorption matrix
    :param esa: excited state absorption matrix
    :param ste: stimulated emission matrix
    :param mpr: multi phonon relaxation matrix
    :param k_etu: vector of overlap integrals for energy transfer upconversion processes
    :param mat_etu: matrix of states for energy transfer upconversion processes
    :param k_cr: vector of overlap integrals for cross relaxation processes
    :param mat_cr: matrix of states for cross relaxation processes
    :param atol: absolute tolerance of inner solver
    :return: list of time steps, list of energy level population vectors
    """
    t0, t1, t_tes = 0, 0.2, 100
    x0 = np.array([1, 0, 0, 0, 0, 0])
    ts = np.linspace(t0, t1, t_tes)
    lin = spe + gsa + esa + ste + mpr
    # Note: Look more into abstol, SUPER performance critical, but also to set low enough to get the correct result.
    xs = scipy.integrate.odeint(uc_ode, x0, ts, atol=atol, args=(lin, k_etu, mat_etu, k_cr, mat_cr))
    return ts, xs


def uc_ode(x, t, lin, k_etu, mat_etu, k_cr, mat_cr):
    """
    Upconversion dynamics (Hofmann2016, eq 7). Details are given in the thesis of Hofmann, see eq 2.33 for example.
    :param x: energy level population vector
    :param t: current time step
    :param lin: matrix of linear terms of rate equation
    :param k_etu: vector of overlap integrals for energy transfer upconversion processes
    :param mat_etu: matrix of states for energy transfer upconversion processes
    :param k_cr: vector of overlap integrals for cross relaxation processes
    :param mat_cr: matrix of states for cross relaxation processes
    :return: stepped energy level population density vector
    """
    vec_etu = np.array([x[1] ** 2.0, x[1] * x[3], x[3] ** 2.0, x[2] ** 2.0, x[3] * x[1],
                        x[4] * x[2], x[2] * x[4], x[2] * x[1], x[3] * x[2]])
    vec_cr = np.array([x[3] * x[0], x[5] * x[0], x[5] * x[1], x[5] * x[0], x[5] * x[0],
                       x[5] * x[1], x[5] * x[1], x[4] * x[0], x[4] * x[1]])
    return np.matmul(lin, x) + np.matmul(vec_etu * mat_etu, k_etu) + np.matmul(vec_cr * mat_cr, k_cr)


def calc_lum_abs(x, einstein, gsa, esa):
    """
    Given the result vector x, i.e. the steady state population, calculate luminescence and absorption.
    :param x: energy level population vector
    :param einstein: matrix of einstein coefficients
    :param gsa: ground state absorption matrix
    :param esa: excited state absorption matrix
    :return: luminescence vector, total absorption
    """

    # Luminescence vector, entry [6] contains 6->2 luminescence (hack), others i contain (i+1) -> 1 luminescence.
    lum = np.zeros(6)
    lum[0:5] = x[1:6] * einstein[0, 1:6]
    lum[5] = x[5] * einstein[1, 5]

    # Absorption, GSA from 1->2 along with ESA from 2->4 and 6->4.
    abs_tot = x[0] * gsa[1, 0] + x[1] * esa[3, 1] + x[3] * esa[5, 3]

    return lum, abs_tot


# endregion


# region Slice building

def default_rem_params(lmbds, irrs, angs=None, ldos_input=None):
    """
    Construction of default parameters for rem. 
    :param lmbds: list of design wavelengths
    :param irrs: list of incident irradiances
    :param angs: list of incident angles (default is normal incidence)
    :param ldos_input: list of rem dos inputs
    :return: dictionary of rem parameters
    """
    return {"ldos_input": [RemDosInput(None, None, None, None, None, None)] if ldos_input is None else ldos_input,
            "lmbd": np.array(lmbds),
            "irr": np.array(irrs),
            "ang": np.array([0]) if angs is None else np.array(angs),
            "lmb": np.array([1523])}


def default_slice_gen(rem_input, res):
    """
    Default generator for slicing rem calculation. The default generator supports varying "ldos_input", "lmbd", "lmb", 
    "ang", "irr". For other needs, a custom slice generator should be created (use this one as a starting point).
    :param rem_input: rem input object ("RemInput")
    :param res: spatial resolution of Wigner Seitz unit cell
    :return: list of parameter names, list of parameter values and list of slices
    """
    # Generate meta data (and check that correct inputs are specified).
    param_names = ["ldos_input", "lmbd", "lmb", "ang", "irr"]
    if set(rem_input.param_map.keys()) != set(param_names):
        raise ValueError("The provided RemInput (specifically, the specified param_map) is not compatible with the "
                         "default slice generator.")
    param_values = [rem_input.param_map[key] if key is not "ldos_input" else
                    [item.name for item in rem_input.param_map[key]] for key in param_names]
    # Generate slices.
    slices = []
    for ldos_input in rem_input.param_map["ldos_input"]:
        # Load LDOS result on-the-fly.
        ldos_input.open()
        for lmbd in rem_input.param_map["lmbd"]:
            # Calculate LDOS correction factor on-the-fly.
            ldos_corr = ldos_correction(np.array(rem_input.params["lmbs"]), ldos_input, lmbd, res) \
                if ldos_input.ldos_rel is not None else [np.ones((6, 6))] * res
            for lmb in rem_input.param_map["lmb"]:
                abs_er = rem_input.abs_spec(lmb) if rem_input.abs_spec is not None else 1
                for ang in rem_input.param_map["ang"]:
                    # Load irradiance enhancement from irr_input.
                    if rem_input.irr_input is not None:
                        irr_brg, irr_ref = rem_input.irr_input.get_irr(lmb, ang, res, lmbd)
                    # If no bragg stack in provided, set the relative irradiance to 1.
                    else:
                        irr_ref = 1
                        irr_brg = np.ones(res)
                    # Inner loop over irradiance (does not affect the optical results).
                    for irr in rem_input.param_map["irr"]:
                        # Physical parameters for setup of REM matrices.
                        params = {"irr": irr, "abs_er": abs_er}
                        params.update(rem_input.params)
                        # Arguments for slice (calculation wise).
                        my_slice = {"params": params, "res": res, "lmbd": lmbd, "irr_ref": irr_ref,
                                    "irr_brg": irr_brg, "ldos_correction": ldos_corr}
                        # Append slices to queue.
                        slices.append(my_slice)
        # Unload LDOS when no more needed.
        ldos_input.close()
    return param_names, param_values, slices

# endregion
