import unittest
import matplotlib.pyplot as plt

from dos_core import *


def compare_dos3_ldos3(path):
    dos_result = load(path)
    dos3d = dos_result.dos3d['tm'] + dos_result.dos3d['te']
    ldos3d = (dos_result.ldos3d['tm'] + dos_result.ldos3d['te']) * dos_result.epsilon
    np.testing.assert_array_almost_equal(dos3d, np.sum(ldos3d, axis=1), 3)


class TestDos(unittest.TestCase):
    def test_sanity(self):
        self.assertTrue(True)

    def test_k_points(self):
        wd = "../tmp_test"
        step = 1e-1
        kx0, kx1 = 0, 0.5
        ky0, ky1 = 0, 2.1
        # Grids to test.
        grid_kx = MpbGrid1D(kx0, kx1, step, 'x', 0)
        grid_ky = MpbGrid1D(ky0, ky1, step, 'y', 0)
        grid_kxy1 = MpbGrid2D(kx0, kx1, ky0, ky1, step, 'y')
        grid_kxy2 = MpbGrid2D(kx0, kx1, ky0, ky1, step, 'x')
        grid_kxy3 = MpbGrid2D(kx0, kx1, ky0, ky1, step)
        # Test parameters.
        mpb_params = {"n_low": 1.5, "n_high": 1.790322580645161, "num_bands": 4, "resolution": 68}
        kx = np.arange(step / 2.0, kx1, step)
        ky = np.arange(step / 2.0, ky1, step)
        # Test 1D grids.
        run_mpb_calculation("kx", MpbInput(mpb_params, grid_kx, field=False), agg=agg_mpb_results, wd=wd)
        result = load(os.path.join(wd, "kx", "mpb_result.json"))
        np.testing.assert_array_almost_equal(result.kx, kx)
        run_mpb_calculation("ky", MpbInput(mpb_params, grid_ky, field=False), agg=agg_mpb_results, cores=1, wd=wd)
        result = load(os.path.join(wd, "ky", "mpb_result.json"))
        np.testing.assert_array_almost_equal(result.ky, ky)
        # Test 2D grid, slicing along y.
        run_mpb_calculation("grid_kxy1", MpbInput(mpb_params, grid_kxy1, field=False), agg=agg_mpb_results, wd=wd)
        result = load(os.path.join(wd, "grid_kxy1", "mpb_result.json"))
        np.testing.assert_array_almost_equal(result.ky, np.array(list(ky) * len(kx)))
        np.testing.assert_array_almost_equal(result.kx, np.repeat(kx, len(ky)))
        # Test 2D grid, slicing along x.
        run_mpb_calculation("grid_kxy2", MpbInput(mpb_params, grid_kxy2, field=False), agg=agg_mpb_results, wd=wd)
        result = load(os.path.join(wd, "grid_kxy2", "mpb_result.json"))
        np.testing.assert_array_almost_equal(result.kx, np.array(list(kx) * len(ky)))
        np.testing.assert_array_almost_equal(result.ky, np.repeat(ky, len(kx)))
        # Test 2D grid, no slicing (result equal to slicing along y!).
        run_mpb_calculation("grid_kxy3", MpbInput(mpb_params, grid_kxy3, field=False), agg=agg_mpb_results,
                            wd="../tmp_test")
        result = load(os.path.join(wd, "grid_kxy3", "mpb_result.json"))
        np.testing.assert_array_almost_equal(result.ky, np.array(list(ky) * len(kx)))
        np.testing.assert_array_almost_equal(result.kx, np.repeat(kx, len(ky)))

    def test_ldos_grid_interp(self):
        wd = "../tmp_test"
        step = 1e-1
        # Initialize mpb objects.
        params = {"n_low": 1.5, "n_high": 1.8, "num_bands": 4, "resolution": 256}
        mpb_grid = MpbGrid2D(0, 0.5, 0, 2.1, step)
        mpb_input = MpbInput(params, mpb_grid)
        args = {"w_max": 1.2, "dw": [2e-3, 1e-3]}
        none = "test_interp_running_no_interp"
        xint = "test_interp_running_x_interp"
        yint = "test_interp_running_y_interp"
        # Run the calculations.
        run_mpb_calculation(none, mpb_input, calc=calc_dos_result, args=args, wd=wd, agg=agg_dos_results, cores=10)
        mpb_grid.mpb_interp = 'x'
        run_mpb_calculation(xint, mpb_input, calc=calc_dos_result, args=args, wd=wd, agg=agg_dos_results, cores=10)
        mpb_grid.mpb_interp = 'y'
        run_mpb_calculation(yint, mpb_input, calc=calc_dos_result, args=args, wd=wd, agg=agg_dos_results, cores=10)
        # Compare results.
        for i in range(len(args["dw"])):
            name = "dos_result_dw{:.0e}.json".format(args["dw"][i])
            self.compare_ldos_results(wd, none, [xint, yint], name)

    def test_mpb_sym(self):
        visual = False
        # Control variables.
        res = 8
        step = 1e-1
        num_bands = 4
        # Names.
        wd = "../tmp_test"
        sym = "test_sym_res{}".format(res)
        nosym = "test_nosym_res{}".format(res)
        # Setup parameters.
        n_low = 1.0
        n_high = 5.0 / 3.0
        # Check that w_high in pixels is an even integer
        w_high = res * n_low / (n_high + n_low)
        self.assertAlmostEqual(w_high % 2, 1)
        # Initialize mpb objects.
        brg_param = {"n_low": n_low, "n_high": n_high, "num_bands": num_bands, "resolution": res}
        mpb_grid = MpbGrid2D(0, 0.5, 0, 2.1, step)
        brg_sym = MpbInput(brg_param, mpb_grid, sym=True)
        brg_nosym = MpbInput(brg_param, mpb_grid)
        # Run calculations.
        run_mpb_calculation(sym, brg_sym, agg=agg_mpb_results, wd=wd)
        run_mpb_calculation(nosym, brg_nosym, agg=agg_mpb_results, wd=wd)
        # Load results.
        result_sym = load(os.path.join(wd, sym, "mpb_result.json"))
        result_nosym = load(os.path.join(wd, nosym, "mpb_result.json"))
        # Test the fields.
        for key in ["te", "tm"]:
            for i in range(result_sym.ws[key].shape[0]):
                for j in range(result_sym.ws[key].shape[1]):
                    recons = result_sym.fields[key][i, j, :]
                    original = result_nosym.fields[key][i, j, :]
                if visual:
                    plt.plot(recons, '--')
                    plt.plot(original)
                np.testing.assert_array_almost_equal(original, recons, 2)
        if visual:
            plt.xlim(0, res - 1)
            plt.show()

    def test_mpb_aggregation(self):
        wd = "../tmp_test"
        run = True
        step = 1e-1
        modes = ["te", "tm"]
        # Initialize mpb objects.
        params = {"n_low": 1.5, "n_high": 1.8, "num_bands": 4, "resolution": 32}
        mpb_grid = MpbGrid2D(0, 0.5, 0, 2.0, step)
        mpb_input = MpbInput(params, mpb_grid, modes)
        # Calculation names.
        ref = "test_mpb_aggregation_ref"
        int = "test_mpb_aggregation_int"
        par = "test_mpb_aggregation_par"
        # Run the calculations.
        if run:
            run_mpb_calculation(ref, mpb_input, agg=agg_mpb_results, cores=1, wd=wd)
            mpb_grid.mpb_interp = 'y'
            run_mpb_calculation(int, mpb_input, agg=agg_mpb_results, cores=1, wd=wd)
            run_mpb_calculation(par, mpb_input, agg=agg_mpb_results, wd=wd)
        # Load results.
        result_ref = load(os.path.join(wd, ref, 'mpb_result.json'))
        result_int = load(os.path.join(wd, int, 'mpb_result.json'))
        result_par = load(os.path.join(wd, par, 'mpb_result.json'))
        # Compare results (field data are only precise to two digits (and varies!), others are precise and fixed).
        precision = {"fields": 2}
        for key in result_ref.__dict__:
            d = 5 if key not in precision else precision[key]
            if result_ref.__dict__[key] is None:
                continue
            if isinstance(result_ref.__dict__[key], dict):
                for subkey in result_ref.__dict__[key]:
                    np.testing.assert_array_almost_equal(result_ref.__dict__[key][subkey],
                                                         result_int.__dict__[key][subkey], d, err_msg=key)
                    np.testing.assert_array_almost_equal(result_ref.__dict__[key][subkey],
                                                         result_par.__dict__[key][subkey], d, err_msg=key)
                continue
            # print("{}:{}\n".format(key, result_ref.__dict__[key]))
            # print("{}:{}\n".format(key, result_int.__dict__[key]))
            # print("{}:{}\n".format(key, result_par.__dict__[key]))
            d = 5 if key not in precision else precision[key]
            np.testing.assert_array_almost_equal(result_ref.__dict__[key], result_int.__dict__[key], d, err_msg=key)
            np.testing.assert_array_almost_equal(result_ref.__dict__[key], result_par.__dict__[key], d, err_msg=key)

    def test_calc_ldos_running(self):
        wd = "../tmp_test"
        step = 1e-1
        # Initialize mpb objects.
        params = {"n_low": 1.5, "n_high": 1.8, "num_bands": 4, "resolution": 32}
        mpb_grid = MpbGrid2D(0, 0.5, 0, 2.0, step, mpb_interp='y')
        mpb_input = MpbInput(params, mpb_grid)
        args = {"w_max": 1.2, "dw": [2e-3, 1e-3]}
        # Calculation names.
        ref = "test_calc_ldos_running_ref"
        run = "test_calc_ldos_running_run"
        chk = "test_calc_ldos_running_chk"
        # Run the calculations.
        run_mpb_calculation(ref, mpb_input, wd=wd)
        run_mpb_calculation(run, mpb_input, calc=calc_dos_result, args=args, wd=wd, agg=agg_dos_results)
        run_mpb_calculation(chk, mpb_input, calc=calc_dos_result, args=args, wd=wd, agg=agg_dos_results,
                            chunk_size=np.inf)
        for i, dw in enumerate(args["dw"]):
            w_bins = np.arange(0, args["w_max"], dw)
            clean_up = i == len(args["dw"]) - 1
            dos_result_from_mpb_result(os.path.join(wd, ref), w_bins, clean_up=clean_up)
        # Check that results are equivalent.
        for i, dw in enumerate(args["dw"]):
            name = "dos_result_dw{:.0e}.json".format(dw)
            self.compare_ldos_results(wd, ref, [run, chk], name)

    def test_calc_frac_ldos(self):
        wd = "../tmp_test"
        step = 1e-1
        # Initialize mpb objects.
        params = {"n_low": 1.5, "n_high": 1.8, "num_bands": 4, "resolution": 32}
        mpb_grid = MpbGrid2D(0, 0.5, 0, 2.0, step, mpb_interp='y')
        mpb_input = MpbInput(params, mpb_grid)
        args = {"w_max": 1.2, "dw": [1e-3]}
        # Run reference calculation.
        ref = "test_calc_frac_ldos_ref"
        run_mpb_calculation(ref, mpb_input, calc=calc_dos_result, args=args, wd=wd, agg=agg_dos_results)
        # Run "normal" calculation.
        norm = "test_calc_frac_ldos_norm"
        args["n_out"] = 1
        args["phi_d"] = [0]
        args["theta_d"] = 0.6
        run_mpb_calculation(norm, mpb_input, calc=calc_frac_dos_result, args=args, wd=wd, agg=agg_dos_results)
        # Run calculation with almost n_out infinite (i.e. equal to the high index layer).
        ninf = "test_calc_frac_ldos_ninf"
        my_args = dict(args)
        my_args["n_out"] = 1e5
        run_mpb_calculation(ninf, mpb_input, calc=calc_frac_dos_result, args=my_args, wd=wd, agg=agg_dos_results)
        # Run calculation with omnidirectional collection.
        omni = "test_calc_frac_ldos_omni"
        my_args = dict(args)
        my_args["theta_d"] = 180
        run_mpb_calculation(omni, mpb_input, calc=calc_frac_dos_result, args=my_args, wd=wd, agg=agg_dos_results)
        # Check that results are equivalent.
        for i, dw in enumerate(args["dw"]):
            for j, phi_d in enumerate(args["phi_d"]):
                name = "dos_result_dw{:.0e}.json".format(dw)
                frac_name = "dos_result_dw{:.0e}_phi{}.json".format(dw, phi_d)
                self.compare_ldos_results(wd, ref, [omni, ninf], name, [frac_name, frac_name])
                self.compare_ldos_results_not(wd, ref, [norm], name, [frac_name])

    def test_compare_resource_ref_dos3d_to_theory(self):
        visual = False
        # Load the LDOS calcuation.
        wd = "res/"
        dos_result = load(os.path.join(wd, "ref_res32", "dos_result_dw1e-03.json"))
        dos3d = dos_result.dos3d['tm'] + dos_result.dos3d['te']
        # Calculate theoretical expression.
        eps = np.min(dos_result.epsilon)
        rho = dos3d_analytic(np.sqrt(eps), dos_result.dk, dos_result.w_bins)
        # Compare.
        if visual:
            plt.plot(dos_result.w_bins, dos3d)
            plt.plot(dos_result.w_bins, rho, ls='--')
            plt.legend()
            plt.show()
        self.assertAlmostEqual(np.mean(rho - dos3d) / 100, 0, 0)

    def test_resource_ldos_normalization(self):
        compare_dos3_ldos3(os.path.join("res/", "ref_res32", "dos_result_dw1e-03.json"))
        compare_dos3_ldos3(os.path.join("res/", "brg_res32", "dos_result_dw1e-03.json"))

    def test_ldos_normalization(self):
        wd = "../tmp_test"
        step = 1e-1
        # Initialize mpb objects.
        params = {"n_low": 1.5, "n_high": 1.8, "num_bands": 4, "resolution": 32}
        mpb_grid = MpbGrid2D(0, 0.5, 0, 2.0, step, mpb_interp='y')
        # Calculation names.
        ref = "test_calc_ldos_running_ref"
        sym = "test_calc_ldos_running_sym"
        # Run dos calc.
        args = {"w_max": 1.2, "dw": [1e-3]}
        mpb_input = MpbInput(params, mpb_grid)
        run_mpb_calculation(ref, mpb_input, calc=calc_dos_result, args=args, wd=wd, agg=agg_dos_results)
        mpb_input = MpbInput(params, mpb_grid, sym=True)
        run_mpb_calculation(sym, mpb_input, calc=calc_dos_result, args=args, wd=wd, agg=agg_dos_results)
        # Compare LDOS sum and DOS.
        name = "dos_result_dw{:.0e}.json".format(1e-3)
        compare_dos3_ldos3(os.path.join(wd, ref, name))
        compare_dos3_ldos3(os.path.join(wd, sym, name))

    # region Utils

    def compare_ldos_results(self, wd, ref, others, name, other_names=None):
        result_ref = load(os.path.join(wd, ref, name))
        for i, entry in enumerate(others):
            other_name = other_names[i] if other_names is not None else name
            other = load(os.path.join(wd, entry, other_name))
            self.assertEqual(result_ref.sym, other.sym)
            np.testing.assert_array_almost_equal(result_ref.w_bins, other.w_bins)
            for key in ["dos2d", "dos3d", "ldos2d", "ldos3d"]:
                for pol in ["tm", "te"]:
                    if key is "ldos2d" and result_ref.__dict__[key] is None:
                        self.assertIsNone(other.__dict__[key])
                        continue
                    np.testing.assert_array_almost_equal(result_ref.__dict__[key][pol],
                                                         other.__dict__[key][pol], 0, err_msg=key)

    def compare_ldos_results_not(self, wd, ref, others, name, other_names=None):
        np.testing.utils.assert_raises(AssertionError, self.compare_ldos_results, wd, ref, others, name, other_names)

        # endregion

