import unittest
import matplotlib.pyplot as plt

from irr_core import *
from tmmpy_mc import *
from tmmpy_utils import *


class TestIrr(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Material values only valid at 1523 nm (toy materials for testing).
        cls.pmma = read_material("pmma")
        cls.pmma_np = read_material("pmma_np")
        cls.tio2 = read_material("tio2")
        cls.lmbd = 1620.5
        cls.num_al = 19
        cls.lmb = 1523
        cls.pol = 's'
        cls.ang = 0
        cls.res = 100

    def test_sanity(self):
        self.assertTrue(True)

    # region Comparison to Hofmann data.

    def test_figure3a(self):
        visual = False
        # Construct variable-resolution design wavelength vector.
        lmbds = list(np.arange(500, 2005, 5)) + list(np.arange(1590, 1640, 0.5)) + list(
            np.arange(1598.5, 1600, 0.1))
        lmbds = np.sort(np.array(lmbds))
        # Calculate mean irradiance as function of design wavelength.
        bs = BraggStack(self.pmma_np, self.tio2, self.lmbd, self.num_al, self.pmma)
        irr_enhancements = calc_mean_rel_irrs(self.lmb, self.pol, self.ang, bs, lmbds)
        # Load Hofmann data.
        hofmann = np.genfromtxt("res/fig3a.csv", delimiter=",")
        hofmann = interp(hofmann[:, 0], hofmann[:, 1], bounds_error=False)(lmbds)
        if visual:
            plt.plot(lmbds, hofmann)
            plt.plot(lmbds, irr_enhancements, '--')
            plt.show()
        # Test stuff.
        np.testing.assert_array_almost_equal(hofmann, irr_enhancements, 2)

    def test_figure3b(self):
        visual = False
        # Calculate data.
        bs = BraggStack(self.pmma_np, self.tio2, self.lmbd, self.num_al, self.pmma)
        res_vec = [[i + 1, self.res] for i in range(bs.num_al * 2 + 3)]
        pts, irr = tc._scan_stack_tmm(bs, self.lmb, self.pol, self.ang, res_vec,
                                      lambda i, pre, E: eval_irr(i, pre, E, 's'))
        # Load Hofmann data.
        hofmann = np.genfromtxt("res/fig3b.csv", delimiter=",")
        hofmann = interp(hofmann[:, 0], hofmann[:, 1], bounds_error=False)(pts)
        # Plotting for debugging.
        if visual:
            plt.plot(pts, hofmann)
            plt.plot(pts, irr, '--')
            plt.show()
        # Test stuff.
        delta = np.mean((hofmann - irr) ** 2)
        self.assertAlmostEqual(delta, 0, 3)

    def test_figure3c(self):
        visual = False
        # Calculate data.
        lmbs = np.linspace(1000, 2000, 1001)
        bs = BraggStack(self.pmma_np, self.tio2, self.lmbd, self.num_al, self.pmma)
        result = solve_tmm(bs, lmbs, 's', 0, seq=True)
        reflectance = result['R']
        # Load Hofmann data.
        hofmann = np.genfromtxt("res/fig3c.csv", delimiter=",")
        hofmann = interp(hofmann[:, 0], hofmann[:, 1])(lmbs)
        # Plotting for debugging.
        if visual:
            plt.plot(lmbs, hofmann)
            plt.plot(lmbs, reflectance, '--')
            plt.show()
        # Test stuff.
        np.testing.assert_array_almost_equal(hofmann, reflectance, 3)

    def test_figure4a(self):
        visual = False
        # Calculate data.
        num_als = np.array([12, 20, 50]) - 1
        lmbds = np.arange(1580, 1700, 0.1)
        irr_map = {}
        for num_al in num_als:
            bs = BraggStack(self.pmma_np, self.tio2, None, num_al, self.pmma)
            irr_enhancements = calc_mean_rel_irrs(self.lmb, self.pol, self.ang, bs, lmbds)
            irr_map[num_al] = irr_enhancements
        # Calculate data.
        hofmann_map = {}
        for num_al in num_als:
            tmp = np.genfromtxt("res/fig4a{}.csv".format(num_al + 1), delimiter=",")
            hofmann_map[num_al] = interp(tmp[:, 0], tmp[:, 1])(lmbds)
        # Plotting for debugging.
        if visual:
            for num_al in num_als:
                plt.plot(lmbds, hofmann_map[num_al])
                plt.plot(lmbds, irr_map[num_al], '--')
            plt.show()
        # Test stuff.
        for num_al in num_als:
            delta = np.mean((hofmann_map[num_al] - irr_map[num_al]) ** 2)
            self.assertAlmostEqual(delta, 0, 3)

    def test_figure4b(self):
        visual = False
        # Calculate data.
        delta = 0.1
        lmbd_map = {3: np.arange(1600, 1800, delta), 7: np.arange(1600, 1700, delta),
                    11: np.arange(1600, 1700, delta)}
        num_als = np.array([4, 8, 12, 16] + list(np.arange(20, 105, 5))) - 1
        irr_map = {}
        max_map = {}
        for num_al in num_als:
            # Default value for range.
            if num_al not in lmbd_map:
                lmbd_map[num_al] = np.arange(1550, 1650, delta)
            # Do the magic.
            lmbds = lmbd_map.get(num_al)
            bs = BraggStack(self.pmma_np, self.tio2, None, num_al, self.pmma)
            irr_enhancements = calc_mean_rel_irrs(self.lmb, self.pol, self.ang, bs, lmbds)
            irr_map[num_al] = irr_enhancements
            max_map[num_al] = np.argmax(irr_enhancements)
        # Prepare data.
        gamma_max = [irr_map[key][max_map[key]] for key in num_als]
        lmbd_max = [lmbd_map[key][max_map[key]] for key in num_als]
        # Load Hofmann data.
        hofmann_gamma = np.genfromtxt("res/fig4b1.csv", delimiter=",")
        hofmann_lmbd = np.genfromtxt("res/fig4b2.csv", delimiter=",")
        if visual:
            plt.plot(hofmann_lmbd[:, 0] - 1, hofmann_lmbd[:, 1], ls='none', marker='o')
            plt.plot(num_als, lmbd_max, ls='none', marker='x', label="Calculated")
            plt.figure()
            plt.plot(hofmann_gamma[:, 0] - 1, hofmann_gamma[:, 1], ls='none', marker='o')
            plt.plot(num_als, gamma_max, ls='none', marker='x')
            plt.show()
        # Test stuff.
        delta = np.mean((hofmann_lmbd[:, 1] - lmbd_max) ** 2)
        self.assertAlmostEqual(delta / 100, 0, 1)
        delta = np.mean((hofmann_gamma[:, 1] - gamma_max) ** 2)
        self.assertAlmostEqual(delta / 10, 0, 1)

    # endregion

    # region Irradiance calculation core

    def test_irr_result(self):
        bs = BraggStack(Material(1.5, 0), Material(4.0, 0), None, 2, Material(1.5, 0))
        # param_map = default_irr_params([1400, 1500, 1600])
        param_map = default_irr_params([1400])
        run_irr_calculation("test_irr_result", bs, param_map, wd="../tmp_test", res=100)
        # Test that pre processed / instant processing agrees.
        simple = RemIrrInput(bs)
        cached = RemIrrFileInput(os.path.join("../tmp_test", "test_irr_result", "irr_result.json"))
        for lmbd in param_map["lmbd"]:
            # Test that cache matches at SAME resolution.
            irr_brg_simple, irr_ref_simple = simple.get_irr(param_map["lmb"][0], param_map["ang"][0], 100, lmbd)
            irr_brg_cached, irr_ref_cached = cached.get_irr(param_map["lmb"][0], param_map["ang"][0], 100, lmbd)
            np.testing.assert_array_almost_equal(irr_brg_simple, irr_brg_cached, 10)
            np.testing.assert_array_almost_equal(irr_ref_simple, irr_ref_cached, 10)
            # Test that cache matches at DIFFERENT resolution (interpolation).
            irr_brg_simple, irr_ref_simple = simple.get_irr(param_map["lmb"][0], param_map["ang"][0], 5, lmbd)
            irr_brg_cached, irr_ref_cached = cached.get_irr(param_map["lmb"][0], param_map["ang"][0], 5, lmbd)
            np.testing.assert_array_almost_equal(irr_brg_simple, irr_brg_cached, 10)
            np.testing.assert_array_almost_equal(irr_ref_simple, irr_ref_cached, 10)

    def test_data_extraction(self):
        # Setup test structure/parameters.
        bs = BraggStack(Material(1.5, 0.0), Material(3.0, 0.0), np.nan, 5)
        lmbs, lmbds, angs = [500, 1000, 1500], [2000, 2500], [0, np.pi / 8, np.pi / 6, np.pi / 4, np.pi / 2]
        param_map = {"lmb": np.array(lmbs), "lmbd": np.array(lmbds), "ang": np.array(angs)}
        # Run calculation.
        irr_result = run_irr_calculation("mwe", bs, param_map)

        def test_select_value(exact):
            for lmb in lmbs:
                for lmbd in lmbds:
                    for ang in angs:
                        fltr = {"lmb": lmb, "lmbd": lmbd, "ang": ang}
                        value = select_value(irr_result, fltr, exact=exact)
                        reference = run_irr_calculation("test", bs, {key: np.array([fltr[key]]) for key in fltr})
                        np.testing.assert_array_almost_equal(value[0], reference.irrs[0][0])

        test_select_value(True)
        test_select_value(False)

    # endregion

    # region Monte Carlo

    def test_mc_zero_std(self):
        bs = BraggStack(Material(1.5, 0), Material(4.0, 0), 1984, 25, Material(1.5, 0))
        # Test with integral.
        mean_rel_irr_enh = calc_mean_rel_irr(1523, 's', 0, bs)
        mean_rel_irr_enh_mc, _ = calc_mean_rel_irr_mc(1523, 's', 0, bs, 5, 1e-12)
        self.assertAlmostEqual(mean_rel_irr_enh, mean_rel_irr_enh_mc)
        # Test with irradiance distribution.
        pts, irr_brg, irr_ref = calc_irr(1523, 0, bs, 16)
        pts, irr_brg_mc, irr_ref_mc = calc_irr_mc(1523, 0, bs, 16, 5, 1e-12)
        np.testing.assert_array_almost_equal(irr_brg, irr_brg_mc)

    # TODO: Add more MC tests

    # endregion

    # region Oblique incidence

    def test_const_irr_for_reference(self):
        lmb, res = 1523, 100
        bs = BraggStack(Material(1.5, 0), Material(4.0, 0), 1984, 25, Material(1.5, 0))
        # Construct reference.
        res_vec = [[1, res * bs.num_al]]
        active_idxs = bs.get_active_idxs()
        d_lst = [l.depth for l in bs.assemble()]
        ref = construct_reference_stack(bs, np.sum([d_lst[idx] for idx in active_idxs]))
        # Calculate stuff.
        for pol in ['s', 'p']:
            for ang in [0, np.pi / 8, np.pi / 4]:
                pts, irr = tc._scan_stack_tmm(ref, lmb, pol, ang, res_vec, lambda i, pre, E: eval_irr(i, pre, E, pol))
                transmission = solve_tmm(ref, lmb, pol, ang, seq=True)['T'][0]
                np.testing.assert_array_almost_equal(irr, np.ones_like(irr) * transmission)

    # TODO: Add much more tests (this part is MUCH more complicated than normal incidence)

    # endregion

    def test_bs_build(self):
        air = Material(1.0, 0)
        for entry in [2, 4, 10, 100]:
            bs = BraggStack(air, air, 500, entry, air)
            layers = bs.assemble()
            self.assertEqual(len(layers), entry * 2 + 5)

    def test_bs_build_no_lmb8(self):
        air = Material(1.0, 0)
        for entry in [2, 4, 10, 100]:
            bs = BraggStack(air, air, 500, entry)
            layers = bs.assemble()
            self.assertEqual(len(layers), entry * 2 + 3)

    def test_irr_integral(self):
        visual = False
        lmb, pol, ang, lmbd = 1532, 's', 0, 1620.5
        pmma, pmma_np, tio2 = Material(1.49, 0), Material(1.48, 0), Material(1.73, 0)
        bs = BraggStack(pmma_np, tio2, lmbd, 0, pmma)
        num_als = [20, 50]
        for num_al in num_als:
            bs.num_al = num_al
            irr_enc = calc_mean_rel_irr(lmb, pol, ang, bs)
            irr_num = []
            res_lst = [2, 5, 10, 20, 50, 100, 500, 2500]
            for res in res_lst:
                irr_num.append(calc_mean_rel_irr_numeric(lmb, pol, ang, bs, res))
            if visual:
                plt.plot(res_lst, [irr_enc] * len(irr_num))
                plt.plot(res_lst, irr_num)
                plt.show()
            self.assertAlmostEqual(irr_enc.real, irr_num[-1], 3)
