import unittest
import matplotlib.pyplot as plt

from rem_core import *
from dos_utils import *
from bragg_utils import *


class TestRem(unittest.TestCase):
    def test_sanity(self):
        self.assertTrue(True)

    def test_compare_matlab(self):
        rem_input = RemInput(hofmann2016(), None, default_rem_params([0], [100, 1000]))
        result = run_rem("test_matlab_compare", rem_input, cores=1, wd="../tmp_test")
        # Check state conservation.
        for entry in result.vec_ref:
            self.assertAlmostEqual(np.sum(entry), 1.0, 10)
        # Vectors calculated from Hoffmanns MATLAB code.
        vec1 = [9.99972495e-01, 2.55231591e-05, 1.93874619e-06, 4.16795403e-08, 1.09351584e-09, 1.95393757e-10]
        vec2 = [9.99996917e-01, 3.04873301e-06, 3.33161238e-08, 5.27616887e-10, 3.17019701e-12, 6.95960058e-14]
        # Compare results.
        np.testing.assert_array_almost_equal(result.vec_ref[0], np.array(vec2))
        np.testing.assert_array_almost_equal(result.vec_ref[1], np.array(vec1))

    def test_compare_benni(self):
        visual = False
        # Data from Benni (i think?). TODO: Update these data!
        fn = "res/140527_9ETU_9CR_Prozesse_g_i_und_g_f_vertauscht_mit_Korrekturfaktor_fuer_ET.csv"
        data = np.loadtxt(fn, delimiter=";")
        # Calculate data, low resolution to limit computation time.
        irrs = (data[:, 0]).tolist()
        params = hofmann2016()
        params["er_doping"] = 0.225
        rem_input = RemInput(params, None, default_rem_params([1523], irrs))
        result = run_rem("test_matlab_compare", rem_input, wd="../tmp_test")
        # Visualization for debugging purposes.
        if visual:
            for i in range(1, len(result.vec_ref[0])):
                plt.plot(irrs, 100 * result.get_qy_ref()[:, i],
                         label="N={}".format(i + 2))
                plt.plot(data[:, 0], data[:, i], ls='--', label="Benni: N={}".format(i + 2))
                # print("Benni: {}".format(data[-1, (i+1)]))
                # print("Emher: {}".format(items[-1]))
            plt.legend()
            plt.show()
        # Compare results.
        for i in range(1, len(result.vec_ref[0])):
            qys = 100 * result.get_qy_ref()[:, i]
            np.testing.assert_array_almost_equal(qys, data[:, i], 1)

    def test_ucqy_max(self):
        wd = "../tmp_test"
        # Load the LDOS calcuation.
        brg_file = os.path.join("res", "brg_res32", "dos_result_dw1e-03.json")
        ldos_input = RemDosFileInput(brg_file, fltr=lambda x: gf1d(x, 3))
        # Build the bragg stack.
        pmma = read_material("pmma")
        pmma_np = read_material("pmma_np")
        tio2 = read_material("tio2")
        # 1) Without irradiance enhancement as in Hofmann fig6.
        rem_input = RemInput(hofmann2016(), None, default_rem_params([1624], [11540], ldos_input=[ldos_input]))
        rem_result = run_rem("test_ucqy_max_fig6", rem_input, wd=wd)
        self.assertAlmostEqual(rem_result.get_ucqy_ref()[0] * 100, 13.89, 2)
        rem_input = RemInput(hofmann2016(), None, default_rem_params([1624], [13620], ldos_input=[ldos_input]))
        rem_result = run_rem("test_ucqy_max_fig6", rem_input, wd=wd)
        self.assertAlmostEqual(rem_result.get_ucqy_brg()[0] * 100, 14.70, 2)
        # 2) With irradiance enhancement as in Hofmann fig7.
        bs = BraggStack(pmma_np, tio2, None, 40, pmma)
        rem_input = RemInput(hofmann2016(), RemIrrInput(bs),
                             default_rem_params([1624], [11540], ldos_input=[ldos_input]))
        rem_result = run_rem("test_ucqy_max_fig7", rem_input, wd=wd)
        self.assertAlmostEqual(rem_result.get_ucqy_ref()[0] * 100, 13.88, 2)
        rem_input = RemInput(hofmann2016(), RemIrrInput(bs),
                             default_rem_params([1624], [3760], ldos_input=[ldos_input]))
        rem_result = run_rem("test_ucqy_max_fig7", rem_input, wd=wd)
        self.assertAlmostEqual(rem_result.get_ucqy_brg()[0] * 100, 14.33, 1)  # low precision due to low ldos resolution
