from numpy.random import normal
from tmmpy_utils import *


def solve_tmm_mc(s, lmb, pol, ang, eval_map=None, itr=1000, std=1, std_map=None):
    """
    Main entry point for solver using the standard transfer matrix method (tmm) with a monte carlo extension. Rather
    than calling the tmm interface of tmmpy, a semi-custom (copy-pasted from tmmpy_core and modified) code is used to
    allow speeding up the calculation utilizing the specific assumptions of the monte carlo run (e.g. the interfaces
    for each run are constant, so they can be pre processed).
    :param s: stack of layers
    :param lmb: wavelength to evaluate
    :param pol: polarization, "s" or "p"
    :param ang: angle of incidence (in rad)
    :param eval_map: custom function to evaluate
    :param itr: number of monte carlo iterations
    :param std: standard deviation of the gaussian distribution from which the d_depths are drawn
    :param std_map: like std, but map of (material name, std) values (to enable different stds for different materials)
    :return: map of output, including per default power reflectance R and power transmittance T
    """
    eval_map = {} if eval_map is None else eval_map
    # Do pre processing.
    pre = pre_process(s, lmb, ang)
    m = len(pre['d'])
    stds = [std_map[pre['layers'][i].material.name] if np.isfinite(pre['d'][i]) else 0 for i in range(m)] if std_map is not None else [std]*m
    # Allocate coherent interface/layer matrices.
    I_lst = np.empty((m - 1, 2, 2), dtype=complex)
    L_lst = np.empty((m - 1, 2, 2), dtype=complex)
    # Calculate interface matrices.
    for i in np.arange(0, m - 1):
        j = i + 1
        # Coherent interface matrix.
        I_lst[i] = coherent_interface_matrix(pre['n'][i], pre['n'][j], pre['ang'][i], pre['ang'][j], pol)

    def func(idx):
        new_d = [pre['d'][i] + normal(scale=stds[i]) for i in range(len(pre['d']))]
        # Calculate coherent propagation matrices.
        for i in np.arange(0, m - 1):
            j = i + 1
            if j < m - 1:
                L_lst[j] = coherent_propagation_matrix(new_d[j] * pre['k'][j])
        # Assemble data.
        data = {'S': tc._assemble_s_matrix(I_lst, L_lst), 'I': I_lst, 'L': L_lst, 'd': new_d}
        # Create default output.
        out = {'r': tc._r_from_S(data['S']), 't': tc._t_from_S(data['S'])}
        out['R'] = tc._R_from_r(out['r']).real
        out['T'] = tc._T_from_t(out['t'], pre['n'][0], pre['n'][-1], pre['ang'][0], pre['ang'][-1], pol).real
        # Create custom output.
        for key in sorted(eval_map.keys()):  # it's a hack to sort them
            out[key] = eval_map[key](data, out)
        return out

    return tc._sweep(func, itr, seq=True)


# region Relative irradiance calculation

def calc_mean_rel_irr_mc(lmb, pol, ang, bs, itr=1000, std=1):
    """
    Calculate mean relative irradiance across the structure, integral evaluated analytically.
    :param lmb: excitation wavelength
    :param pol: polarisation "s" or "p"
    :param ang: angle of incidence (in rad)
    :param bs: bragg stack object ("BraggStack")
    :param itr: number of monte carlo iterations
    :param std: standard deviation of the gaussian distribution from which the d_depths are drawn
    :return: mean relative irradiance
    """
    pre = pre_process(bs, lmb, ang)
    # Calculate integral.
    result = solve_tmm_mc(bs, lmb, pol, ang, eval_map={'E': eval_E_coherent}, itr=itr, std=std)
    d_tot = np.sum([pre['d'][i] for i in bs.get_active_idxs()])
    integrals = np.array(
        [np.sum([eval_irr_int(pre, result['E'][j], i) for i in bs.get_active_idxs()]) for j in range(itr)])
    # Calculate reference integral.
    bs_ref = construct_reference_stack(bs, d_tot)
    result_ref = solve_tmm(bs_ref, lmb, pol, ang, eval_map={'E': eval_E_coherent}, seq=True)
    integral_ref = eval_irr_int(pre_process(bs_ref, lmb, ang), result_ref['E'][0], 1)
    return np.mean(integrals / integral_ref), np.std(integrals / integral_ref)


def calc_irr_mc(lmb, ang, bs, res=100, itr=1000, std=1):
    """
    Main entry point for calculation of the relative irradiance across the structure with monte carlo extension as used
    as input into the rem.
    :param lmb: excitation wavelength
    :param ang: angle of incidence (in rad)
    :param bs: bragg stack object ("BraggStack")
    :param res: spatial resolution per layer
    :param itr: number of monte carlo iterations
    :param std: standard deviation of the gaussian distribution from which the d_depths are draw
    :return: vector of relative irradiance at each point in the active layers, polarisation averaged
    """
    return calc_irr(lmb, ang, bs, res,
                    lambda lmb, ang, bs, res, pol: calc_irr_inner_mc(lmb, ang, bs, res, pol, itr, std))


def calc_irr_inner_mc(lmb, ang, bs, res, pol, itr=1000, std=1):
    """
    Calculate relative irradiance across the structure with monte carlo extension. Rather
    than calling the tmm interface of tmmpy, a semi-custom (copy-pasted from tmmpy_core and modified) code is used to
    allow speeding up the calculation utilizing the specific assumptions of the monte carlo run (e.g. the interfaces
    for each run are constant, so they can be pre processed).
    :param lmb: excitation wavelength
    :param ang: angle of incidence (in rad)
    :param bs: bragg stack object ("BraggStack")
    :param res: spatial resolution per layer
    :param pol: polarisation "s" or "p"
    :param itr: number of monte carlo iterations
    :param std: standard deviation of the gaussian distribution from which the d_depths are draw
    :return: vector of spatial points, vector of relative irradiance at each point in the active layers
    """
    pre = pre_process(bs, lmb, ang)
    active_idxs = bs.get_active_idxs()
    res_vec = [[i, res] for i in active_idxs]
    # Allocate beta matrix for speed.
    beta_matrix = array([[exp(1j * 0), 0], [0, exp(- 1j * 0)]], dtype=complex)

    def inner_func(beta, i, out):
        beta_matrix[0, 0] = exp(1j * beta)
        beta_matrix[1, 1] = exp(-1j * beta)
        return eval_irr(i, pre, np.dot(beta_matrix, out['E'][i]), pol)

    def scan_func(data, out):
        pre['d'] = data['d']
        pts, scan = tc._scan_stack(pre, lambda beta, i: inner_func(beta, i, out), res_vec)
        return scan

    # Calculate bragg points Monte Carlo style.
    result = solve_tmm_mc(bs, lmb, pol, ang, {'E': eval_E_coherent, 'irr_scan': scan_func}, itr, std)
    irr_mc = np.mean(np.array(result['irr_scan']), axis=0)
    # Calculate reference points.
    bs_ref = construct_reference_stack(bs, np.sum([pre['d'][idx] for idx in active_idxs]))
    irr_ref = solve_tmm(bs_ref, lmb, pol, ang, seq=True)['T'][0]
    pts, _ = tc._scan_stack(pre, lambda x, y: 0, res_vec)
    return pts, irr_mc, irr_ref

# endregion
