import tmmpy.tmmpy_core as tc

from tmmpy import *


# region Irradiance evaluation

def eval_irr(i, pre, E, pol):
    """
    Evaluate irradiance from electric field amplitudes of forward and backward propagating waves.
    :param i: layer index
    :param pre: pre processing dict ("pre_process" in tmmpy)
    :param E: electric field amplitude, E[0] forward, E[1] backward
    :param pol: polarisation "s" or "p"
    :return: irradiance
    """
    return np.sum(abs(E_vec_from_E(E, pol, pre['ang'][i])) ** 2) * \
           tc._gamma(pre['n'][0], pre['n'][i], pre['ang'][0], pre['ang'][i], pol)


# TODO: Update with angle and polarization dependence (sit down and derive it).
def eval_irr_int(pre, E, i):
    """
    Evaluate irradiance integral analytically from electric field amplitudes of forward and backward propagating waves
    (much faster/more accurate than numerical integration).
    :param i: layer index
    :param pre: pre processing dict ("pre_process" in tmmpy)
    :param E: electric field amplitude, E[0] forward, E[1] backward
    :return: integral irradiance across layer i
    """
    if pre['ang'][0] > 0:
        raise NotImplementedError("eval_irr_int supports only normal incidence evaluation")
    v, w = E[i][0], E[i][1]
    n, d, k = pre['n'][i], pre['d'][i], pre['k'][i]
    term1 = conj(n) * d * (abs(v) ** 2 + abs(w) ** 2)
    term2 = conj(n) / (2.0 * 1.0j * k) * \
            ((v * conj(w)) * (exp(2.0 * 1.0j * k * d) - 1) - (conj(v) * w) * (exp(-2.0 * 1.0j * k * d) - 1))
    return (term1 + term2).real


# endregion

# region Energy density/irradiance across the stack

def scan_irr(bs, lmb, pol, ang, res):
    """
    Local irradiance in all layers of a bragg stack
    :param bs: bragg stack object ("BraggStack")
    :param lmb: excitation wavelength
    :param pol: polarisation "s" or "p"
    :param ang: angle of incidence (in rad)
    :param res: spatial resolution in each layer
    :return: vector of spatial points, vector of local irradiance values
    """
    add_layers = 3 if bs.lmb8 is not None else 1
    res_vec = [[i + 1, res] for i in range(bs.num_al * 2 + add_layers)]
    return tc._scan_stack_tmm(bs, lmb, pol, ang, res_vec, lambda i, pre, E: eval_irr(i, pre, E, pol))

# endregion


# region Mean relative irradiance calculation

def calc_mean_rel_irrs(lmb, pol, ang, bs, lmbds):
    """
    Wrapper for easy scanning over lmbds.
    :param lmb: excitation wavelength
    :param pol: polarisation "s" or "p"
    :param ang: angle of incidence (in rad)
    :param bs: bragg stack object ("BraggStack")
    :param lmbds: vector of design wavelengths
    :return: vector of mean relative irradiance across the active layers
    """
    design_res = len(lmbds)
    irr_enhancements = np.zeros(design_res)
    for i in range(design_res):
        bs.lmbd = lmbds[i]
        irr_enhancements[i] = calc_mean_rel_irr(lmb, pol, ang, bs)
    return irr_enhancements


def calc_mean_rel_irr(lmb, pol, ang, bs):
    """
    Calculate mean relative irradiance across the structure, integral evaluated analytically.
    :param lmb: excitation wavelength
    :param pol: polarisation "s" or "p"
    :param ang: angle of incidence (in rad)
    :param bs: bragg stack object ("BraggStack")
    :return: mean relative irradiance across the active layers
    """
    pre = pre_process(bs, lmb, ang)
    # Calculate integral.
    result = solve_tmm(bs, lmb, pol, ang, eval_map={'E': eval_E_coherent}, seq=True)
    d_tot = np.sum([pre['d'][i] for i in bs.get_active_idxs()])
    integral = np.sum([eval_irr_int(pre, result['E'][0], i) for i in bs.get_active_idxs()])
    # Calculate reference integral.
    bs_ref = construct_reference_stack(bs, d_tot)
    result_ref = solve_tmm(bs_ref, lmb, pol, ang, eval_map={'E': eval_E_coherent}, seq=True)
    integral_ref = eval_irr_int(pre_process(bs_ref, lmb, ang), result_ref['E'][0], 1)
    return integral / integral_ref


def calc_mean_rel_irr_numeric(lmb, pol, ang, bs, res=100):
    """
    Calculate mean relative irradiance across the structure, integral evaluated numerically. Used only for testing
    purposes.
    :param lmb: excitation wavelength
    :param pol: polarisation "s" or "p"
    :param ang: angle of incidence (in rad)
    :param bs: bragg stack object ("BraggStack")
    :param res: number of points evaluated per layer
    :return: mean relative irradiance across the active layers
    """
    active_idxs = bs.get_active_idxs()
    d_lst = [l.depth for l in bs.assemble()]
    # Calculate integral.
    res_vec = [[i, res] for i in active_idxs]
    pts, irr = tc._scan_stack_tmm(bs, lmb, pol, ang, res_vec, lambda i, pre, E: eval_irr(i, pre, E, 's'))
    d_tot = 0.0
    integral = 0
    for i in range(len(active_idxs)):
        x = pts[i * res:(i + 1) * res]
        integral += np.trapz(irr[i * res:(i + 1) * res], x=x)
        d_tot += d_lst[active_idxs[i]]  # (x[-1] - x[0])
    # Calculate reference integral.
    bs_ref = construct_reference_stack(bs, d_tot)
    integral_ref = d_tot * solve_tmm(bs_ref, lmb, pol, ang, seq=True)['T'][0]
    return integral / integral_ref


# endregion

# region Spatially resolved relative irradiance

def calc_irr(lmb, ang, bs, res=100, inner=None):
    """
    Main entry point for calculation of the relative irradiance across the structure, polarisation averaged.
    :param lmb: excitation wavelength
    :param ang: angle of incidence (in rad)
    :param bs: bragg stack object ("BraggStack")
    :param res: number of points evaluated per layer
    :param inner: relative irradiance calculation function (default is "calc_rel_irr_inner", for mc simulations use
    "calc_rel_irr_inner_mc")
    :return: vector of spatial points, vector of relative irradiance at each point in the active layers
    """
    inner = calc_irr_inner if inner is None else inner
    # No bragg stack, no enhancement.
    if bs is None:
        return np.zeros(1), np.ones(1)
    # Normal incidence, polarization average not needed.
    if ang == 0:
        return inner(lmb, ang, bs, res, 's')
    # Non-normal incidence, do polarization average.
    pts, sbrg, sref = inner(lmb, ang, bs, res, 's')
    pts, pbrg, pref = inner(lmb, ang, bs, res, 'p')
    return pts, (sbrg + pbrg) / 2.0, (sref + pref) / 2.0


# TODO: The new approach of evaluating the reference is FAST, but it ONLY makes sense if it's non-absorbing
def calc_irr_inner(lmb, ang, bs, res, pol):
    """
    Calculate irradiance across the structure.
    :param lmb: excitation wavelength
    :param ang: angle of incidence (in rad)
    :param bs: bragg stack object ("BraggStack")
    :param res: number of points evaluated per layer
    :param pol: polarisation "s" or "p"
    :return: vector of spatial points, vector of relative irradiance at each point in the active layers
    """
    active_idxs = bs.get_active_idxs()
    d_lst = [l.depth for l in bs.assemble()]
    # Calculate points.
    res_vec = [[i, res] for i in active_idxs]
    pts, irr = tc._scan_stack_tmm(bs, lmb, pol, ang, res_vec, lambda i, pre, E: eval_irr(i, pre, E, pol))
    # Calculate reference points.
    bs_ref = construct_reference_stack(bs, np.sum([d_lst[idx] for idx in active_idxs]))
    irr_ref = solve_tmm(bs_ref, lmb, pol, ang, seq=True)['T'][0]
    return pts, np.array(irr).real, irr_ref


# endregion

def construct_reference_stack(bs, d_tot):
    """
    Given a bragg stack, constructs the corresponding reference stack.
    :param bs: bragg stack object ("BraggStack")
    :param d_tot: total thickness of active layers (not calculated from Bragg stack for mc performance reasons,
    see the mc code)
    :return: reference stack object ("Stack", see tmmpy code)
    """
    bs_ref = Stack()
    bs_ref.left = bs.left
    bs_ref.append(Layer(bs.low, d_tot))
    bs_ref.right = bs.right
    return bs_ref
