import json
import os
import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy

from dos_utils import *
from tmmpy_utils import *
from scipy.interpolate import interp1d as interp

jsonpickle_numpy.register_handlers()

# The __data_version__ property makes it possible to support import of legacy data in the future (so far not used).
__data_version__ = 1


# region Objects for mpb interface


class MpbInput:
    def __init__(self, mpb_params, mpb_grid, modes=None, field=None, sym=False):
        """
        Object holding all inputs relevant to the mpb calculation.
        :param mpb_params: parameters which are printed directly in the .ctl script. These should contain 
        "n_low", "n_high", "num_bands", "resolution".
        :param mpb_grid: the grid of k-point which is evaluated, can be both 1D ("MpbGrid1D") and 2D ("MpbGrid2D")
        :param modes: which modes to evaluate, default is to include both te and tm
        :param field: which field to evaluate (should typically NOT be changed)
        :param sym: of True, symmetry of the fields in real space is exploited to save only half of the fields
        """
        self.mpb_params = mpb_params
        self.mpb_grid = mpb_grid
        self.modes = modes if modes is not None else ["te", "tm"]
        self.field = None if field is False else (field if field is not None else "dpwr")
        self.sym = sym


class MpbGrid2D:
    def __init__(self, kx_min, kx_max, ky_min, ky_max, dk, mpb_interp=None, offset=True):
        """
        Object holding all inputs relevant to the construction of a 2D grid for the MPB calculation.
        :param kx_min: minimum kx value
        :param kx_max: maximum kx value
        :param ky_min: minimum ky value
        :param ky_max: maximum ky value
        :param dk: spacing in kx, ky (only uniform grid spacing is supported)
        :param mpb_interp: axis for which interpolation is carried out in mpb, must be x or y. Typically the y 
        direction is the best choice
        :param offset: if True (default), k-points are offset by 1/2 step from the origin
        """
        self.dk = dk
        self.ky_max = ky_max
        self.ky_min = ky_min + (self.dk / 2.0 if offset else 0)
        self.kx_max = kx_max
        self.kx_min = kx_min + (self.dk / 2.0 if offset else 0)
        self.mpb_interp = mpb_interp

    def get_slices(self):
        """
        Calculations are typically divided into slices. This functionality must be supported by the grid object.
        :return: 2d grid in k-space-slices for mpb calculation
        """
        kxs = np.arange(self.kx_min, self.kx_max, self.dk)
        kys = np.arange(self.ky_min, self.ky_max, self.dk)
        # Interpolation in mpb is in x, interpolate here in y.
        if self.mpb_interp is 'x':
            return [{"x1": kxs[0], "y1": ky, "x2": kxs[-1], "y2": ky, "num-ks": len(kxs) - 2} for ky in kys]
        # Interpolation in mpb is in y, interpolate here in x.
        elif self.mpb_interp is 'y':
            return [{"x1": kx, "y1": kys[0], "x2": kx, "y2": kys[-1], "num-ks": len(kys) - 2} for kx in kxs]
        # No interpolation in mpb, i.e. both interpolations are to be done in python.
        return [{"x1": kxs[0], "y1": kys[0], "x2": kxs[-1], "y2": kys[-1], "dx": self.dk, "dy": self.dk}]


class MpbGrid1D:
    def __init__(self, ki_min, ki_max, dk, axis, k_const, offset=True):
        """
        Object holding all inputs relevant to the construction of a 1D grid for the MPB calculation.
        :param ki_min: minimum ki value
        :param ki_max: maximum ki value
        :param dk: spacing in ki
        :param axis: the axis along with the k-points runs, must be x or y
        :param k_const: the constant k-coordinate, e.g. if axis is x this is the (constant) y-coordinate
        :param offset: if True (default), k-points are offset by 1/2 step from the origin
        """
        if axis not in ['x', 'y']:
            raise ValueError("Axis must be x or y.")
        self.dk = dk
        self.axis = axis
        self.kj_const = k_const
        self.ki_min = ki_min + (self.dk / 2.0 if offset else 0)
        self.ki_max = ki_max - (self.dk / 2.0 if offset else 0)

    def get_slices(self):
        """
        Calculations are typically divided into slices. This functionality must be supported by the grid object.
        :return: 1d grid in k space slices for mpb calculation
        """
        res = (self.ki_max - self.ki_min) / self.dk
        if self.axis == 'x':
            return [{"x1": self.ki_min, "y1": self.kj_const, "x2": self.ki_max, "y2": self.kj_const, "num-ks": res - 1}]
        elif self.axis == 'y':
            return [{"x1": self.kj_const, "y1": self.ki_min, "x2": self.kj_const, "y2": self.ki_max, "num-ks": res - 1}]


# endregion

# region Objects for rem interface

class RemInput:
    def __init__(self, params, irr_input, param_map, abs_spec=None):
        """
        Simple object collecting the inputs needed for running the rate equation model.
        :param params: physical parameter set, e.g. "hofmann2016()"
        :param irr_input: irradiance input ("RemIrrInput")
        :param param_map: parameters over which slices are created. The actual content depends on the slice generator
        used but could be e.g. "irr" for irradiance. For a concrete example see "default_slice_gen", an example of
        arguments is available in "default_rem_params"
        :param abs_spec: absorption spectrum (interpolation object). The default behaviour is to assume monochromatic
        excitation
        """
        self.params = params
        self.param_map = param_map
        self.abs_spec = abs_spec
        self.irr_input = irr_input


class RemDosInput:
    def __init__(self, name, w_bins, ldos_rel, n_low, h_high, info):
        """
        Simple object fulfilling (or defining?) the "interface" for providing LDOS data for the REM calculation.
        :param name: short name describing the underlying LDOS calculation, used only for info purposes (meta data)
        :param w_bins: vector of omega bins for LDOS calculation
        :param ldos_rel: relative LDOS (3D)
        :param n_low: refractive index of low index layer
        :param h_high: refractive indes of high index layer
        :param info: information string
        """
        self.name = name
        self.w_bins = w_bins
        self.n_high = h_high
        self.n_low = n_low
        self.ldos_rel = ldos_rel
        self._info = info  # private property, should NOT be accessed directly

    def get_info(self):
        """
        Collect info on how the object was created and return it.
        :return: info string
        """
        return self._info

    def open(self):
        """
        Load/initialize the LDOS objects. Might be expensive/memory intensive.
        :return: None
        """
        pass

    def close(self):
        """
        Free any resources that were allocted on the open() call.
        :return: None
        """
        pass


class RemDosFileInput:
    def __init__(self, dos_result_brg_file, fltr=None):
        """
        Object allowing LDOS input into REM calculations from files. For method documentation see interface definition
        ("RemDosInput"). On initialization, only the meta data are read. The actual LDOS read (and processing) is
        delayed to the open() call and discarded by the close() call.
        :param dos_result_brg_file: file containing the Bragg structure dos result ("DosResult" object in .json format)
        :param fltr: filter to apply for smoothing out binning noise of the LDOS. Typical choice is a gaussian filter
        """
        self.name = "{}".format(os.path.basename(dos_result_brg_file))
        self.dos_result_brg_file = dos_result_brg_file
        self.fltr = fltr
        # Properties to be set up later.
        self.n_low, self.n_high, self.w_bins, self.ldos_rel = None, None, None, None

    def get_info(self):
        info = {"dos_result_brg_file": self.dos_result_brg_file,
                "fltr": self.fltr.__name__ if self.fltr is not None else ""}
        # Gather all possible info on the dos results.
        with open(os.path.join(os.path.dirname(self.dos_result_brg_file), "mpb.info"), 'r') as in_file:
            info["dos_result_brg_info"] = json.load(in_file)
        return info

    def open(self):
        dos_result_brg = load(self.dos_result_brg_file)
        # Set the relevant parameters.
        self.n_low = np.sqrt(np.min(dos_result_brg.epsilon))
        self.n_high = np.sqrt(np.max(dos_result_brg.epsilon))
        self.w_bins = dos_result_brg.w_bins
        # Calculate LDOS3D polarization average.
        epsilon = dos_result_brg.epsilon
        ldos3d_rel = assemble_dos_data(dos_result_brg, "ldos3d", rel=True, flip=False, fltr=self.fltr)
        low_part = ldos3d_rel[:, abs(epsilon - min(epsilon)) < abs(epsilon - max(epsilon))]
        # low_part = ldos3d_rel[:, abs(epsilon - min(epsilon)) <= abs(epsilon - max(epsilon))]
        low_center = int(np.ceil(low_part.shape[1] / 2))
        self.ldos_rel = np.concatenate((low_part[:, low_center:], low_part[:, :low_center]), 1)

    def close(self):
        self.n_low = None
        self.n_high = None
        self.w_bins = None
        self.ldos_rel = None


class RemIrrInput:
    def __init__(self, bs):
        """
        Simple object fulfilling (or defining?) the "interface" for providing IRR data for the REM calculation. Here IRR
        data refers to the local irradiance (or energy density) relative to the incident irradiance.
        :param bs: bragg stack ("BraggStack" object)
        """
        self.bs = bs

    def get_info(self):
        """
        Collect info on how the object was created and return it.
        :return: info string
        """
        return jsonpickle.encode(self.bs)

    def get_irr(self, lmb, ang, res, lmbd=None):
        """
        Get (spatially resolved) irradiance enhancement.
        :param lmb: excitation wavelength
        :param ang: angle of incidence
        :param res: spatial resolution, total number of points in all active layers
        :param lmbd: design wavelength
        :return: irradiance enhancement (vector)
        """
        self.bs.lmbd = lmbd if lmbd is not None else self.bs.lmbd
        pts, irr_brg, irr_ref = calc_irr(lmb, ang, self.bs, res)
        return irr_brg, irr_ref


class RemIrrFileInput:
    def __init__(self, irr_result_file):
        """
        Object allowing IRR input from files. For method documentation see interface definition ("RemIrrInput").
        :param irr_result_file: file containing the Bragg structure irr result ("IrrResult" object in .json format)
        """
        self.irr_result_file = irr_result_file
        self.irr_result = load(irr_result_file)

    def get_info(self):
        info = {"irr_result_file": self.irr_result_file}
        # Gather all possible info on the irr result.
        with open(os.path.join(os.path.dirname(self.irr_result_file), "irr.info"), 'r') as in_file:
            info["irr_result_info"] = json.load(in_file)
        return info

    def get_irr(self, lmb, ang, res, lmbd=None):
        arg_map = {"lmb": lmb, "ang": ang, "lmbd": lmbd}
        # Index magic, find the correct entry.
        idx, m = 0, len(self.irr_result.param_values)
        for i in range(m):
            value = arg_map[self.irr_result.param_names[i]]
            if value is None:
                continue
            tmp = self.irr_result.param_values[i].index(value)
            idx += tmp * (len(self.irr_result.param_values[i + 1]) if i < m - 1 else 1)
        irr_brg, irr_ref = self.irr_result.irrs[idx]
        # Interpolate if needed.
        num_al = self.irr_result.props["num_al"]
        if abs(res * num_al - len(irr_brg)) < 0:
            return self.irr_result.irrs[idx]
        return interp(np.linspace(0, 1, len(irr_brg), False), irr_brg, fill_value='extrapolate', bounds_error=False)(
            np.linspace(0, 1, res * num_al, False)), irr_ref


# endregion

# region Objects for tmm interface

class BraggStack:
    def __init__(self, low, high, lmbd, num_al, lmb8=None):
        """
        Class that represents a bragg stack. Mimics a "Stack" object (from tmmpy), so it must fulfill the interface.
        :param low: low index material ("Material" object)
        :param high: high index material ("Material" object)
        :param lmbd: design wavelength
        :param num_al: number of active layers
        :param lmb8: lmb8 layer material ("Material" object). Default behavior is not to add lmb8 layers
        """
        self.low = low
        self.high = high
        self.lmb8 = lmb8
        self.lmbd = lmbd
        self.num_al = num_al
        # Per default, the stack is surrounded by (perfect) vacuum and deposited on a glass substrate.
        self.left = Material(1, 0)
        self.right = Material(1.5, 0)

    def assemble(self):
        """
        Create a layer stack from the Bragg structure information (used inside tmmpy).
        :return: stack of layers (see "Layer" in tmmpy)
        """
        # Add surrounding layers.
        bs = [Layer(self.left, inf)]
        # Build the bragg stack.
        if self.lmb8 is not None:
            bs.append(Layer(self.lmb8, self.lmbd / 8.0 / self.lmb8.get_ri(self.lmbd)))
        for i in range(self.num_al + 1):
            bs.append(Layer(self.high, self.lmbd / (4.0 * self.high.get_ri(self.lmbd))))
            if i < self.num_al:
                bs.append(Layer(self.low, self.lmbd / (4.0 * self.low.get_ri(self.lmbd))))
        if self.lmb8 is not None:
            bs.append(Layer(self.lmb8, self.lmbd / 8.0 / self.lmb8.get_ri(self.lmbd)))
        # Add surrounding layers.
        bs.append(Layer(self.right, inf))
        return bs

    def get_active_idxs(self):
        """
        Get position indices of the active layers.
        :return: list of position indices
        """
        offset = 2 if self.lmb8 is None else 3
        return list(np.arange(self.num_al) * 2 + offset)

    def invert(self):
        """
        Part of the "Stack" object interface, so far not implemented (not necessary for simple TMM calculations).
        :return: None
        """
        raise NotImplementedError()

    def clone(self):
        """
        Clone, i.e. create a deep copy.
        :return: clone
        """
        bs = BraggStack(self.low.clone(), self.high.clone(), self.lmbd, self.num_al, self.lmb8)
        bs.left = self.left.clone()
        bs.right = self.right.clone()
        return bs


# endregion

# region Result objects

class MpbResult(object):
    def __init__(self, kx, ky, kz, epsilon, ws, fields, sym):
        """
        Result object from mpb calculation. Holds all output. Let n denote the number of k-points, m the number of bands
        and l the number of points (pixels) in real space.
        :param kx: kx coordinates (n x 1)
        :param ky: ky coordinates (n x 1)
        :param kz: kz coordinates (n x 1)
        :param epsilon: epsilon function in space (l x 1)
        :param ws: dict of modes (te, tm), each entry holds band frequencies (n x m)
        :param fields: dict of modes (te, tm), each entry holds fields (n x m x l)
        :param sym: if True, symmetry of the fields in real space is exploited to save only half of the field
        """
        self.kx = kx
        self.ky = ky
        self.kz = kz
        self.ws = ws
        self.fields = fields
        self.epsilon = epsilon
        self.sym = sym

    def get_modes(self):
        """
        Get the modes, typically ["te", "tm"].
        :return: list of modes
        """
        return list(self.ws.keys())

    def get_num_bands(self):
        """
        Get the number of bands.
        :return: number of bands
        """
        return self.ws[self.get_modes()[0]].shape[1]

    def reconstruct_from_symmetry(self):
        """
        Reconstuction of spatially symmetric fields.
        :return: None
        """
        if not self.sym:
            return
        self.sym = False
        self.epsilon = reconstruct(self.epsilon)
        for pol in self.fields:
            shape = self.fields[pol].shape
            tmp = np.empty((shape[0], shape[1], shape[2] * 2 - 2))
            for i in range(shape[0]):
                for j in range(shape[1]):
                    tmp[i, j, :] = reconstruct(self.fields[pol][i, j])
            self.fields[pol] = tmp


class DosResult(object):
    def __init__(self, w_bins, dos2d, dos3d, ldos2d, ldos3d, epsilon, dk, sym):
        """
        Result object from mpb calculation. Holds all output. Let n denote the number of points (pixels) in real space
        and m the number of omega bins.
        :param w_bins: angular frequency bins (m x 1)
        :param dos2d: dict of modes (te, tm), each entry holds 2D DOS (m x 1)
        :param dos3d: dict of modes (te, tm), each entry holds 3D DOS (m x 1)
        :param ldos2d: dict of modes (te, tm), each entry holds 2D LDOS (n x m)
        :param ldos3d: dict of modes (te, tm), each entry holds 3D LDOS (n x m)
        :param epsilon: epsilon for structure (n x 1)
        :param sym: if True, symmetry of the LDOS in real space is exploited to save only half of the LDOS
        """
        self.epsilon = epsilon
        self.ldos3d = ldos3d
        self.ldos2d = ldos2d
        self.dos3d = dos3d
        self.dos2d = dos2d
        self.w_bins = w_bins
        self.sym = sym
        self.dk = dk

    def reconstruct_from_symmetry(self):
        """
        Reconstuction of spatially symmetric fields.
        :return: None
        """
        if not self.sym:
            return
        self.sym = False
        self.epsilon = reconstruct(self.epsilon)
        for pol in ["te", "tm"]:
            if self.ldos2d is not None:
                self.ldos2d[pol] = np.array([reconstruct(row) for row in self.ldos2d[pol]])
            self.ldos3d[pol] = np.array([reconstruct(row) for row in self.ldos3d[pol]])

    def clone(self):
        """
        Clone, i.e. create a deep copy.
        :return: clone
        """
        tmp = {} if self.ldos2d is not None else None
        result = DosResult(np.copy(self.w_bins), {}, {}, tmp, {}, np.copy(self.epsilon), self.dk, self.sym)
        for pol in ["te", "tm"]:
            result.dos2d[pol] = np.copy(self.dos2d[pol])
            result.dos3d[pol] = np.copy(self.dos3d[pol])
            if self.ldos2d is not None:
                result.ldos2d[pol] = np.copy(self.ldos2d[pol])
            result.ldos3d[pol] = np.copy(self.ldos3d[pol])
        return result


class RemResult(object):
    def __init__(self, param_names, param_values, rem_entries):
        """
        The results are stored in vectors (vec_ref, lum_ref, ...) where the zero-dimension length is equal to the
        number of parameter combinations investigated. If more than one parameter has been varied, the result vectors
        can/should be reshaped according to the order of the parameters (i.e. the order is IMPORTANT!).
        :param param_names: names of the parameters (design wavelength, irradiance, ...)
        :param param_values: parameter values (lists)
        :param rem_entries: calculates REM entries, i.e. the calculated data
        """
        self.param_names = param_names
        self.param_values = param_values
        self.vec_ref = np.array([item[0] for item in rem_entries])  # vector of population densities of reference
        self.lum_ref = np.array([item[1] for item in rem_entries])  # vector of luminescence of reference
        self.abs_ref = np.array([item[2] for item in rem_entries])  # vector of absorption of reference
        self.vec_brg = np.array([item[3] for item in rem_entries])  # vector of population densities of bragg stack
        self.lum_brg = np.array([item[4] for item in rem_entries])  # vector of luminescence of bragg stack
        self.abs_brg = np.array([item[5] for item in rem_entries])  # vector of absorption of bragg stack

    def get_qy_ref(self):
        """
        Convenient way to get the QY for the reference structure.
        :return: matrix of QYs, each row contains the QY per-transition (see "calc_lum_abs" in "rem_core.py")
        """
        with np.errstate(divide='ignore', invalid='ignore'):
            qy = (self.lum_ref.T / self.abs_ref).T
        qy[np.isnan(qy)] = 0
        return qy

    def get_qy_brg(self):
        """
        Convenient way to get the QY for the Bragg structure.
        :return: matrix of QYs, each row contains the QY per-transition (see "calc_lum_abs" in "rem_core.py")
        """
        with np.errstate(divide='ignore', invalid='ignore'):
            qy = (self.lum_brg.T / self.abs_brg).T
        qy[np.isnan(qy)] = 0
        return qy

    def get_ucqy_ref(self):
        """
        Convenient way to get the UCQY for the reference structure. Note that ONLY the 3->1 transition is included.
        :return: vector of UCQYs
        """
        return np.sum(self.get_qy_ref()[:, 1:], axis=1)

    def get_ucqy_brg(self):
        """
        Convenient way to get the UCQY for the Bragg structure. Note that ONLY the 3->1 transition is included.
        :return: vector of UCQYs
        """
        return np.sum(self.get_qy_brg()[:, 1:], axis=1)

    def get_lum_rel(self):
        """
        Convenient way to get the relative luminescence.
        :return: matrix of lum, each row contains the lum per-transition (see "calc_lum_abs" in "rem_core.py")
        """
        lum_rel = (self.lum_brg / self.lum_ref)
        lum_rel[np.isnan(lum_rel)] = 0
        return lum_rel


class IrrResult(object):
    def __init__(self, param_names, param_values, props, irrs):
        """
        Result object from irradiance calculation. Holds all output.
        :param param_names: names of the parameters, i.e. excitation wavelength "lmb", angle of incidence "ang", spatial
         resolution "res" and design wavelength "lmbd" (list of strings)
        :param param_values: parameter values (list of vectors)
        :param props: dict of properties of the evaluated stack (currently includes number of active layers "num_al")
        :param irrs: irradiance enhancement
        """
        self.param_names = param_names
        self.param_values = param_values
        self.props = props
        self.irrs = irrs


# endregion

# region Result object manipulation logic

# TODO: Add some reshaping logic here? Or indexing logic, similar to

def select_value(result, fltr, exact=None):
    """
    Select a particular "value" (which can be any type of object, not just a scalar) from a result object subject to a
    filtering condition. The current implementation assumes that the filter specifies ALL parameters, and that the
    result yields ONE value. In any other cases, the code might break.
    :param result: the result object from which the data is extracted
    :param fltr: dictionary of (parameter, value) pairs, e.g. {"lmb": 1523, ...}
    :param exact: if True (default) an exact match is required, if False the closest match is returned
    :return: the "value" (which can be an arbitrary object)
    """
    idx = 0
    exact = True if exact is None else exact
    for i, param in enumerate(result.param_names):
        i_idx = result.param_values[i].index(fltr[param]) if exact else \
            np.argmin(abs(np.array(result.param_values[i]) - fltr[param]))
        dim_offset = np.prod([len(lst) for lst in result.param_values[i + 1:]])
        idx += int(i_idx * dim_offset)
    return result.irrs[idx]


def get_param(obj, param):
    """
    Convenient way to extract parameters from e.g. rem results ("RemResult").
    :param obj: the result object
    :param param: the parameter to extract
    :return: parameter values
    """
    return obj.param_values[obj.param_names.index(param)]


# endregion

# region Loading logic

def load(in_path):
    """
    Load object from path. Assumed jsonpickle format and reconstructs from symmetry.
    :param in_path: path of file to load
    :return: loaded object
    """
    with open(in_path, 'r') as in_file:
        result = jsonpickle.decode(in_file.read())
    if hasattr(result, "reconstruct_from_symmetry"):
        result.reconstruct_from_symmetry()
    return result


def reconstruct(field, sym=True):
    """
    Load object from path. Assumed jsonpickle format and reconstructs from symmetry.
    :param field: field to reconstruct
    :param sym: if True, field has been compressed by symmetry
    :return: recontructed field, i.e. without symmetry compression
    """
    if not sym:
        return field
    return np.array(list(field) + list(field[1:-1][::-1]))

# endregion
